import sys

from module_qc_analysis_tools._version import __version__

if sys.version_info >= (3, 9):
    from importlib import resources
else:
    import importlib_resources as resources
data = resources.files("module_qc_analysis_tools") / "data"

__all__ = ("__version__", "data")



#!/usr/bin/env python3
import logging
import sys

from module_qc_analysis_tools.utils.misc import bcolors, getImuxMap, getVmuxMap

log = logging.getLogger(__name__)
log.setLevel("INFO")


def format_text():
    return " {:^30}: {:^20}: {:^20}: {:^5}"


def print_output_pass(key, results, lower_bound, upper_bound):
    txt = format_text()
    log.info(
        bcolors.OKGREEN
        + txt.format(
            key,
            round(results, 4),
            f"[{lower_bound}, {upper_bound}]",
            "PASS",
        )
        + bcolors.ENDC
    )


def print_output_fail(key, results, lower_bound, upper_bound):
    txt = format_text()
    log.info(
        bcolors.BADRED
        + txt.format(
            key,
            round(results, 4),
            f"[{lower_bound}, {upper_bound}]",
            "FAIL",
        )
        + bcolors.ENDC
    )


def get_layer(layer):
    layers = {"L0": "LZero", "L1": "LOne", "L2": "LTwo"}
    return layers.get(layer)


def check_layer(layer):
    possible_layers = ["L0", "L1", "L2"]
    if layer not in possible_layers:
        log.error(
            bcolors.ERROR
            + f" Layer '{layer}' not recognized or not provided. Provide the layer with the --layer [L0, L1, or L2] option."
            + bcolors.ENDC
        )
        sys.exit(-1)


# The following hardcoded values are from https://gitlab.cern.ch/atlas-itk/pixel/module/itkpix-electrical-qc
def get_nominal_current(layer):
    check_layer(layer)
    # Assumes triplets for L0, quads for L1-L2
    currents_per_chip = {
        "L0": 1.85 * 3,
        "L1": 1.65 * 4,
        "L2": 1.47 * 4,
    }
    return currents_per_chip.get(layer)


def get_nominal_Voffs(layer, lp_mode=False):
    check_layer(layer)
    Voffs = {
        "L0": 1.1,
        "L1": 1.0,
        "L2": 1.0,
    }
    Voffs_lp = {
        "L0": 1.38,
        "L1": 1.38,
        "L2": 1.38,
    }
    if lp_mode:
        return Voffs_lp.get(layer)
    else:
        return Voffs.get(layer)


def get_nominal_RextA(layer):
    check_layer(layer)
    RextA = {
        "L0": 511,
        "L1": 732,
        "L2": 866,
    }
    return RextA.get(layer)


def get_nominal_RextD(layer):
    check_layer(layer)
    RextD = {"L0": 407, "L1": 549, "L2": 590}
    return RextD.get(layer)


# Function to get key from value in muxMaps
def get_key(mydict, val):
    for key, value in mydict.items():
        if val == value:
            return key
    return -1


def perform_qc_analysis_AR(test_type, qc_config, layer_name, results, verbosity="INFO"):
    log.setLevel(verbosity)
    # Basically the same function as perform_qc_analysis
    # minus all the initial checks which have already been performed.

    passes_qc_overall = True
    VmuxMap = getVmuxMap()
    ImuxMap = getImuxMap()
    if len(VmuxMap.keys()) + len(ImuxMap.keys()) != len(results):
        log.error(
            bcolors.ERROR
            + " Number of entries in AR_NOMINAL_SETTINGS results does not match number of entries in VmuxMap and ImuxMap - there should be one entry for every Vmux and Imux in those maps. Please fix and re-run!"
            + bcolors.ENDC
        )
        return -1
    for key, value in qc_config.items():

        log.debug(f" QC selections for {key}: {value}")
        try:
            lower_bound = value[0]
            upper_bound = value[1]
        except Exception:
            log.error(
                bcolors.ERROR
                + f" QC selections for {key} are ill-formatted, should be list of length 2! Please fix: {value} . Skipping."
                + bcolors.ENDC
            )
            continue

        passes_qc_test = True
        if get_key(ImuxMap, key) != -1:
            index = get_key(ImuxMap, key)
        elif get_key(VmuxMap, key) != -1:
            index = get_key(VmuxMap, key) + len(ImuxMap.keys())
        else:
            log.error(
                bcolors.ERROR
                + f"Did not find {key} in VmuxMap or ImuxMap - please check!"
                + bcolors.ENDC
            )
            continue
        log.debug(f" Key is {key}, value is {value}, index is {index}")

        if (results[index] < lower_bound) or (results[index] > upper_bound):
            passes_qc_test = False
        if passes_qc_test:
            print_output_pass(key, results[index], lower_bound, upper_bound)
        else:
            print_output_fail(key, results[index], lower_bound, upper_bound)
    passes_qc_overall = passes_qc_overall and passes_qc_test

    return passes_qc_overall


def perform_qc_analysis(test_type, qc_config, layer_name, results, verbosity="INFO"):
    log.setLevel(verbosity)
    log.info("")
    log.info("Performing QC analysis!")
    log.info("")

    qc_selections = qc_config[test_type]
    check_qc_selections = qc_selections.copy()

    check_layer(layer_name)
    layer = get_layer(layer_name)

    passes_qc_overall = True
    txt = format_text()
    log.info(txt.format("Parameter", "Analysis result", "QC criteria", "Pass"))
    log.info(
        "--------------------------------------------------------------------------------------"
    )

    for key in results.keys():

        if not qc_selections.get(key):
            log.warning(
                bcolors.WARNING
                + f" Selection for {key} not found in QC file! Skipping."
                + bcolors.ENDC
            )
            continue
        check_qc_selections.pop(key)

        # Handle AR_NOMINAL_SETTINGS in completely different function, for now...
        if key == "AR_NOMINAL_SETTINGS":
            passes_qc_test = perform_qc_analysis_AR(
                test_type,
                qc_selections.get(key),
                layer_name,
                results.get(key),
                verbosity,
            )
            passes_qc_overall = passes_qc_overall and passes_qc_test
            continue

        log.debug(f" QC selections for {key}: {qc_selections.get(key)}")
        if type(qc_selections.get(key)) is list:
            if len(qc_selections.get(key)) != 2:
                log.error(
                    bcolors.ERROR
                    + f" QC selections for {key} are ill-formatted, should be list of length 2! Please fix: {qc_selections.get(key)} . Skipping."
                    + bcolors.ENDC
                )
                continue
            lower_bound = qc_selections.get(key)[0]
            upper_bound = qc_selections.get(key)[1]
        elif type(qc_selections.get(key)) is dict:
            layer_bounds = qc_selections.get(key).get(layer)
            if not layer_bounds:
                log.error(
                    bcolors.ERROR
                    + f" QC selections for {key} and {layer} do not exist - please check! Skipping."
                    + bcolors.ENDC
                )
                continue
            lower_bound = layer_bounds[0]
            upper_bound = layer_bounds[1]

        passes_qc_test = True
        if ("AR_VDDA_VS_TRIM" in key) or ("AR_VDDD_VS_TRIM" in key):
            # All values in list must satisfy requirements
            for elem in results.get(key):
                tmp_pass_qc = True
                if (elem < lower_bound) or (elem > upper_bound):
                    passes_qc_test = False
                    tmp_pass_qc = False
                if tmp_pass_qc:
                    print_output_pass(key, elem, lower_bound, upper_bound)
                else:
                    print_output_fail(key, elem, lower_bound, upper_bound)

        else:
            if (results.get(key) < lower_bound) or (results.get(key) > upper_bound):
                passes_qc_test = False
            if passes_qc_test:
                print_output_pass(key, results.get(key), lower_bound, upper_bound)
            else:
                print_output_fail(key, results.get(key), lower_bound, upper_bound)
        passes_qc_overall = passes_qc_overall and passes_qc_test
    if len(check_qc_selections) > 0:
        for key in check_qc_selections.keys():
            log.error(
                bcolors.ERROR
                + f" Parameter from chip for QC selection of {key} was not passed to analysis - please fix!"
                + bcolors.ENDC
            )
        passes_qc_overall = False
    log.info(
        "--------------------------------------------------------------------------------------"
    )

    return passes_qc_overall


def submit_results(outputDF, timestamp, site="Unspecified", outputfile="submit.txt"):
    results = outputDF.get("results")

    AR_VDDA_VS_TRIM = 0
    AR_VDDD_VS_TRIM = 0
    if outputDF.get("testType") == "ANALOG_READBACK":
        AR_VDDA_VS_TRIM = sum(results.get("AR_VDDA_VS_TRIM")) / len(
            results.get("AR_VDDA_VS_TRIM")
        )
        AR_VDDD_VS_TRIM = sum(results.get("AR_VDDD_VS_TRIM")) / len(
            results.get("AR_VDDD_VS_TRIM")
        )

    # Temporary solution to avoid error when indexing array that doesn't exist
    if not results.get("AR_NOMINAL_SETTINGS"):
        results.update({"AR_NOMINAL_SETTINGS": [-1] * 72})

    url = {
        "ADC_CALIBRATION": f"https://docs.google.com/forms/d/e/1FAIpQLSegDYRQ1Foe5eTuSOVZUXe0d1f_Bh5v3rhsffCnu9DUDFR69A/formResponse?usp=pp_url\
	&entry.1920584355={timestamp}\
	&entry.104853658={outputDF.get('serialNumber')}\
	&entry.802167553={site}\
	&entry.1592726943={results.get('ADC_CALIBRATION_SLOPE')}\
	&entry.422835427={results.get('ADC_CALIBRATION_OFFSET')}",
        "VCAL_CALIBRATION": f"https://docs.google.com/forms/d/e/1FAIpQLSenLUdLpaHLssp-jdUf1YvqiWvR8WOAkhrpQgfBlZYTWWRNog/formResponse?usp=pp_url\
	&entry.1920584355={timestamp}\
	&entry.104853658={outputDF.get('serialNumber')}\
	&entry.802167553={site}\
	&entry.1592726943={results.get('VCAL_MED_SLOPE')}\
	&entry.422835427={results.get('VCAL_MED_OFFSET')}\
	&entry.424316677={results.get('VCAL_MED_SLOPE_SMALL_RANGE')}\
	&entry.2055663117={results.get('VCAL_MED_OFFSET_SMALL_RANGE')}\
	&entry.1630084203={results.get('VCAL_HIGH_SLOPE')}\
	&entry.1107555352={results.get('VCAL_HIGH_OFFSET')}\
	&entry.1994936328={results.get('VCAL_HIGH_SLOPE_SMALL_RANGE')}\
	&entry.524584120={results.get('VCAL_HIGH_OFFSET_SMALL_RANGE')}",
        "INJECTION_CAPACITANCE": f"https://docs.google.com/forms/d/e/1FAIpQLSfHpq9pjuzgYvjUU8ZHapzCOrIHzyJx3xirJunGEtBO2COYGw/formResponse?usp=pp_url\
	&entry.1920584355={timestamp}\
	&entry.104853658={outputDF.get('serialNumber')}\
	&entry.802167553={site}\
	&entry.1714546984={results.get('INJ_CAPACITANCE')}",
        "SLDO": f"https://docs.google.com/forms/d/e/1FAIpQLSf3NC84OaYYjJ-DgQ29RvMV2dDQnUI0nxBFdnCUVMby7RXOFQ/formResponse?usp=pp_url\
	&entry.910646842={outputDF.get('serialNumber')}\
	&entry.507508481={site}\
	&entry.1425106615={timestamp}\
	&entry.613380586={results.get('SLDO_VI_SLOPE')}\
	&entry.2009791679={results.get('SLDO_VI_OFFSET')}\
	&entry.1877869140={results.get('SLDO_NOM_INPUT_CURRENT')}\
	&entry.1380637801={results.get('SLDO_VDDA')}\
	&entry.959013471={results.get('SLDO_VDDD')}\
	&entry.427742248={results.get('SLDO_VINA')}\
	&entry.1100117192={results.get('SLDO_VIND')}\
	&entry.411324334={results.get('SLDO_VOFFS')}\
	&entry.257023545={results.get('SLDO_IINA')}\
	&entry.172777573={results.get('SLDO_IIND')}\
	&entry.2138863081={results.get('SLDO_IREF')}\
	&entry.1216431295={results.get('SLDO_ISHUNTA')}\
	&entry.825886502={results.get('SLDO_ISHUNTD')}\
	&entry.298426805={results.get('SLDO_ANALOG_OVERHEAD')}\
	&entry.187142037={results.get('SLDO_DIGITAL_OVERHEAD')}",
        "ANALOG_READBACK": f"https://docs.google.com/forms/d/e/1FAIpQLScsfVAnZokYd-CDef1WZGgdNEY-AdqeS3erRF1mzy6Bl37eYg/formResponse?usp=pp_url\
	&entry.910646842={outputDF.get('serialNumber')}\
	&entry.507508481={site}\
	&entry.1425106615={timestamp}\
	&entry.613380586={results.get('AR_NOMINAL_SETTINGS')[0]}\
	&entry.2091108240={results.get('AR_NOMINAL_SETTINGS')[1]}\
	&entry.1308096676={results.get('AR_NOMINAL_SETTINGS')[2]}\
	&entry.1616657488={results.get('AR_NOMINAL_SETTINGS')[3]}\
	&entry.303689355={results.get('AR_NOMINAL_SETTINGS')[4]}\
	&entry.1299197252={results.get('AR_NOMINAL_SETTINGS')[5]}\
	&entry.337124367={results.get('AR_NOMINAL_SETTINGS')[6]}\
	&entry.539725220={results.get('AR_NOMINAL_SETTINGS')[7]}\
	&entry.174520567={results.get('AR_NOMINAL_SETTINGS')[8]}\
	&entry.2077557631={results.get('AR_NOMINAL_SETTINGS')[9]}\
	&entry.1152177529={results.get('AR_NOMINAL_SETTINGS')[10]}\
	&entry.1170074988={results.get('AR_NOMINAL_SETTINGS')[11]}\
	&entry.1695410680={results.get('AR_NOMINAL_SETTINGS')[12]}\
	&entry.1683989630={results.get('AR_NOMINAL_SETTINGS')[13]}\
	&entry.637795568={results.get('AR_NOMINAL_SETTINGS')[14]}\
	&entry.1796334891={results.get('AR_NOMINAL_SETTINGS')[15]}\
	&entry.1192471500={results.get('AR_NOMINAL_SETTINGS')[16]}\
	&entry.1037413000={results.get('AR_NOMINAL_SETTINGS')[17]}\
	&entry.1731827348={results.get('AR_NOMINAL_SETTINGS')[18]}\
	&entry.1788264831={results.get('AR_NOMINAL_SETTINGS')[19]}\
	&entry.1271298835={results.get('AR_NOMINAL_SETTINGS')[20]}\
	&entry.294928269={results.get('AR_NOMINAL_SETTINGS')[21]}\
	&entry.1752002697={results.get('AR_NOMINAL_SETTINGS')[22]}\
	&entry.1789768564={results.get('AR_NOMINAL_SETTINGS')[23]}\
	&entry.19338211={results.get('AR_NOMINAL_SETTINGS')[24]}\
	&entry.1373225730={results.get('AR_NOMINAL_SETTINGS')[25]}\
	&entry.1288561285={results.get('AR_NOMINAL_SETTINGS')[26]}\
	&entry.993587744={results.get('AR_NOMINAL_SETTINGS')[27]}\
	&entry.1225105463={results.get('AR_NOMINAL_SETTINGS')[28]}\
	&entry.2014795413={results.get('AR_NOMINAL_SETTINGS')[29]}\
	&entry.814046228={results.get('AR_NOMINAL_SETTINGS')[30]}\
	&entry.1206599091={results.get('AR_NOMINAL_SETTINGS')[31]}\
	&entry.1046023025={results.get('AR_NOMINAL_SETTINGS')[32]}\
	&entry.125849508={results.get('AR_NOMINAL_SETTINGS')[33]}\
	&entry.278665318={results.get('AR_NOMINAL_SETTINGS')[34]}\
	&entry.1317511634={results.get('AR_NOMINAL_SETTINGS')[35]}\
	&entry.799431715={results.get('AR_NOMINAL_SETTINGS')[36]}\
	&entry.1032356051={results.get('AR_NOMINAL_SETTINGS')[37]}\
	&entry.206739602={results.get('AR_NOMINAL_SETTINGS')[38]}\
	&entry.47441728={results.get('AR_NOMINAL_SETTINGS')[39]}\
	&entry.887166253={results.get('AR_NOMINAL_SETTINGS')[40]}\
	&entry.290527652={results.get('AR_NOMINAL_SETTINGS')[41]}\
	&entry.1481344879={results.get('AR_NOMINAL_SETTINGS')[42]}\
	&entry.155322339={results.get('AR_NOMINAL_SETTINGS')[43]}\
	&entry.556597681={results.get('AR_NOMINAL_SETTINGS')[44]}\
	&entry.1293797041={results.get('AR_NOMINAL_SETTINGS')[45]}\
	&entry.1984481605={results.get('AR_NOMINAL_SETTINGS')[46]}\
	&entry.1633430606={results.get('AR_NOMINAL_SETTINGS')[47]}\
	&entry.1430993123={results.get('AR_NOMINAL_SETTINGS')[48]}\
	&entry.526213623={results.get('AR_NOMINAL_SETTINGS')[49]}\
	&entry.1631275305={results.get('AR_NOMINAL_SETTINGS')[50]}\
	&entry.975590254={results.get('AR_NOMINAL_SETTINGS')[51]}\
	&entry.1474828103={results.get('AR_NOMINAL_SETTINGS')[52]}\
	&entry.1495459865={results.get('AR_NOMINAL_SETTINGS')[53]}\
	&entry.1128496051={results.get('AR_NOMINAL_SETTINGS')[54]}\
	&entry.367477458={results.get('AR_NOMINAL_SETTINGS')[55]}\
	&entry.1466626922={results.get('AR_NOMINAL_SETTINGS')[56]}\
	&entry.631124052={results.get('AR_NOMINAL_SETTINGS')[57]}\
	&entry.946981503={results.get('AR_NOMINAL_SETTINGS')[58]}\
	&entry.571213202={results.get('AR_NOMINAL_SETTINGS')[59]}\
	&entry.688702844={results.get('AR_NOMINAL_SETTINGS')[60]}\
	&entry.431853336={results.get('AR_NOMINAL_SETTINGS')[61]}\
	&entry.1724286670={results.get('AR_NOMINAL_SETTINGS')[62]}\
	&entry.2112361286={results.get('AR_NOMINAL_SETTINGS')[63]}\
	&entry.1689766951={results.get('AR_NOMINAL_SETTINGS')[64]}\
	&entry.2142543004={results.get('AR_NOMINAL_SETTINGS')[65]}\
	&entry.1946421005={results.get('AR_NOMINAL_SETTINGS')[66]}\
	&entry.707341702={results.get('AR_NOMINAL_SETTINGS')[67]}\
	&entry.1328302698={results.get('AR_NOMINAL_SETTINGS')[68]}\
	&entry.1022788500={results.get('AR_NOMINAL_SETTINGS')[69]}\
	&entry.973739200={results.get('AR_NOMINAL_SETTINGS')[70]}\
	&entry.1279705270={results.get('AR_NOMINAL_SETTINGS')[71]}\
	&entry.1637225517={results.get('AR_TEMP_NTC')}\
	&entry.1793217377={results.get('AR_TEMP_EXT')}\
	&entry.2135558015={results.get('AR_TEMP_ASLDO')}\
	&entry.1505388309={results.get('AR_TEMP_DSLDO')}\
	&entry.363112736={results.get('AR_TEMP_ACB')}\
	&entry.1942035528={results.get('AR_TEMP_ACB')}\
	&entry.1251896209={AR_VDDA_VS_TRIM}\
	&entry.1435921809={AR_VDDD_VS_TRIM}",
    }
    log.info(
        bcolors.WARNING
        + "Copy the following URL into a browser to submit these results: \n"
        + url.get(outputDF.get("testType")).replace("\t", "")
        + "\n"
        + "View submitted results at: https://docs.google.com/spreadsheets/d/1pw_07F94fg2GJQr8wlvhaRUV63uhsAuBt_S1FEFBzBU/view"
        + bcolors.ENDC
    )
    outfile = open(outputfile, "a")
    outfile.writelines(url.get(outputDF.get("testType")).replace("\t", "") + "\n")
    outfile.close()



"""
Top-level entrypoint for the command line interface.
"""
from __future__ import annotations

import typer

import module_qc_analysis_tools
from module_qc_analysis_tools.cli.ADC_CALIBRATION import main as adc_calibration
from module_qc_analysis_tools.cli.ANALOG_READBACK import main as analog_readback
from module_qc_analysis_tools.cli.globals import CONTEXT_SETTINGS
from module_qc_analysis_tools.cli.INJECTION_CAPACITANCE import (
    main as injection_capacitance,
)
from module_qc_analysis_tools.cli.overwrite_config import main as overwrite_config
from module_qc_analysis_tools.cli.SLDO import main as sldo
from module_qc_analysis_tools.cli.update_chip_config import main as update_chip_config
from module_qc_analysis_tools.cli.VCAL_CALIBRATION import main as vcal_calibration

# subcommands
app = typer.Typer(context_settings=CONTEXT_SETTINGS)
app_analysis = typer.Typer(context_settings=CONTEXT_SETTINGS)
app_config = typer.Typer(context_settings=CONTEXT_SETTINGS)
app.add_typer(app_analysis, name="analysis")
app.add_typer(app_config, name="config")


@app.callback(invoke_without_command=True)
def main(
    version: bool = typer.Option(False, "--version", help="Print the current version."),
    prefix: bool = typer.Option(
        False, "--prefix", help="Print the path prefix for data files."
    ),
) -> None:
    """
    Manage top-level options
    """
    if version:
        typer.echo(f"module-qc-analysis-tools v{module_qc_analysis_tools.__version__}")
        raise typer.Exit()
    if prefix:
        typer.echo(module_qc_analysis_tools.data.resolve())
        raise typer.Exit()


app_analysis.command("adc-calibration")(adc_calibration)
app_analysis.command("analog-readback")(analog_readback)
app_analysis.command("injection-capacitance")(injection_capacitance)
app_analysis.command("sldo")(sldo)
app_analysis.command("vcal-calibration")(vcal_calibration)
app_config.command("overwrite")(overwrite_config)
app_config.command("update")(update_chip_config)

# for generating documentation using mkdocs-click
typer_click_object = typer.main.get_command(app)



from __future__ import annotations

import logging
import os
from datetime import datetime
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
import typer
from module_qc_data_tools import load_json, outputDataFrame, qcDataFrame, save_dict_list

from module_qc_analysis_tools.cli.globals import (
    CONTEXT_SETTINGS,
    OPTIONS,
    FitMethod,
    LogLevel,
)
from module_qc_analysis_tools.utils.analysis import perform_qc_analysis, submit_results
from module_qc_analysis_tools.utils.misc import (
    DataExtractor,
    JsonChecker,
    bcolors,
    get_inputs,
    get_qc_config,
    get_time_stamp,
    linear_fit,
    linear_fit_np,
)

app = typer.Typer(context_settings=CONTEXT_SETTINGS)


@app.command()
def main(
    input_meas: Path = OPTIONS["input_meas"],
    base_output_dir: Path = OPTIONS["output_dir"],
    qc_criteria_path: Path = OPTIONS["qc_criteria"],
    layer: str = OPTIONS["layer"],
    permodule: bool = OPTIONS["permodule"],
    submit: bool = OPTIONS["submit"],
    site: str = OPTIONS["site"],
    fit_method: FitMethod = OPTIONS["fit_method"],
    verbosity: LogLevel = OPTIONS["verbosity"],
):
    log = logging.getLogger(__name__)
    log.setLevel(verbosity.value)

    if submit and site == "":
        log.error(
            bcolors.ERROR
            + "You have supplied the --submit option without specifying --site (testing site). Please specify your testing site if you would like to submit these results"
            + bcolors.ENDC
        )
        return

    log.info("")
    log.info(" ===============================================")
    log.info(" \tPerforming ADC calibration analysis")
    log.info(" ===============================================")
    log.info("")

    test_type = os.path.basename(__file__).split(".py")[0]

    time_start = round(datetime.timestamp(datetime.now()))
    output_dir = base_output_dir.joinpath(test_type).joinpath(f"{time_start}")
    output_dir.mkdir(parents=True, exist_ok=False)

    allinputs = get_inputs(input_meas, test_type)
    qc_config = get_qc_config(qc_criteria_path, test_type)

    alloutput = []
    timestamps = []
    for filename in sorted(allinputs):
        log.info("")
        log.info(f" Loading {filename}")
        meas_timestamp = get_time_stamp(filename)
        inputDFs = load_json(filename)
        log.debug(
            f" There are results from {len(inputDFs)} chip(s) stored in this file"
        )
        for inputDF in inputDFs:
            qcframe = inputDF.get_results()
            metadata = qcframe.get_meta_data()

            """"" Check file integrity  """ ""
            checker = JsonChecker(inputDF, test_type)

            try:
                checker.check()
            except BaseException as exc:
                log.exception(exc)
                log.warning(
                    bcolors.WARNING
                    + " JsonChecker check not passed, skipping this input."
                    + bcolors.ENDC
                )
                continue
            else:
                log.debug(" JsonChecker check passed!")
                pass

            try:
                chipname = metadata.get("Name")
                log.debug(f" Found chip name = {chipname} from chip config")
            except Exception:
                log.warning(
                    bcolors.WARNING
                    + "Chip name not found in input from {filename}, skipping."
                    + bcolors.ENDC
                )
                continue

            """""  Calculate quanties   """ ""
            # Vmux conversion is embedded.
            extractor = DataExtractor(inputDF, test_type)
            calculated_data = extractor.calculate()

            """""        Plotting       """ ""
            x_key = "ADC_Vmux8"
            x = calculated_data.pop(x_key)
            y_key = "VcalMed"

            value = calculated_data.get(y_key)
            if fit_method.value == "root":
                p1, p0 = linear_fit(x["Values"], value["Values"])
            if fit_method.value == "numpy":
                p1, p0 = linear_fit_np(x["Values"], value["Values"])
            # Convert from V to mV
            p1mv = p1 * 1000
            p0mv = p0 * 1000
            fig, ax1 = plt.subplots()
            ax1.plot(
                x["Values"],
                value["Values"],
                "o",
                label="Measured data",
                markersize=10,
            )
            x_line = np.linspace(x["Values"][0], x["Values"][-1], 100)
            ax1.plot(x_line, p1 * x_line + p0, "r--", label="Fitted line")
            ax1.text(
                x["Values"][0],
                0.75 * value["Values"][-1],
                f"y = {p1:.4e} * x + {p0:.4e}",
            )
            ax1.set_xlabel(f"{x_key}[{x['Unit']}]")
            ax1.set_ylabel(f"{y_key}[{value['Unit']}]")
            ax1.set_title(chipname)
            ax1.legend()
            outfile = output_dir.joinpath(f"{chipname}.png")
            log.info(f" Saving {outfile}")
            fig.savefig(f"{outfile}")

            # Load values to dictionary for QC analysis
            results = {}
            results.update({"ADC_CALIBRATION_SLOPE": round(p1mv, 3)})
            results.update({"ADC_CALIBRATION_OFFSET": round(p0mv, 3)})

            # Perform QC analysis
            passes_qc = perform_qc_analysis(
                test_type, qc_config, layer, results, verbosity.value
            )
            if passes_qc == -1:
                log.error(
                    bcolors.ERROR
                    + f" QC analysis for {chipname} was NOT successful. Please fix and re-run. Continuing to next chip.."
                    + bcolors.ENDC
                )
                continue
            log.info("")
            if passes_qc:
                log.info(
                    f" Chip {chipname} passes QC? "
                    + bcolors.OKGREEN
                    + f"{passes_qc}"
                    + bcolors.ENDC
                )
            else:
                log.info(
                    f" Chip {chipname} passes QC? "
                    + bcolors.BADRED
                    + f"{passes_qc}"
                    + bcolors.ENDC
                )
            log.info("")

            """"" Output a json file """ ""
            outputDF = outputDataFrame()
            outputDF.set_test_type(test_type)
            data = qcDataFrame()
            data._meta_data.update(metadata)
            data.add_property(
                "ANALYSIS_VERSION",
                pkg_resources.get_distribution("module-qc-analysis-tools").version,
            )
            for key, value in results.items():
                data.add_parameter(key, value)
            outputDF.set_results(data)
            outputDF.set_pass_flag(passes_qc)
            if submit:
                submit_results(
                    outputDF.to_dict(True),
                    time_start,
                    site,
                    output_dir.joinpath("submit.txt"),
                )
            if permodule:
                alloutput += [outputDF.to_dict(True)]
                timestamps += [meas_timestamp]
            else:
                outfile = output_dir.joinpath(f"{chipname}.json")
                log.info(f" Saving output of analysis to: {outfile}")
                save_dict_list(outfile, [outputDF.to_dict(True)])
    if permodule:
        # Only store results from same timestamp into same file
        dfs = np.array(alloutput)
        tss = np.array(timestamps)
        for x in np.unique(tss):
            outfile = output_dir.joinpath("module.json")
            log.info(f" Saving output of analysis to: {outfile}")
            save_dict_list(
                outfile,
                dfs[tss == x].tolist(),
            )


if __name__ == "__main__":
    typer.run(main)



from __future__ import annotations

import logging
import os
from datetime import datetime
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pkg_resources
import typer
from module_qc_data_tools import load_json, outputDataFrame, qcDataFrame, save_dict_list

from module_qc_analysis_tools.cli.globals import (
    CONTEXT_SETTINGS,
    OPTIONS,
    FitMethod,
    LogLevel,
)
from module_qc_analysis_tools.utils.analysis import (
    get_nominal_current,
    get_nominal_RextA,
    get_nominal_RextD,
    get_nominal_Voffs,
    perform_qc_analysis,
    submit_results,
)
from module_qc_analysis_tools.utils.misc import (
    DataExtractor,
    JsonChecker,
    bcolors,
    get_inputs,
    get_qc_config,
    get_time_stamp,
    linear_fit,
    linear_fit_np,
)

app = typer.Typer(context_settings=CONTEXT_SETTINGS)


@app.command()
def main(
    input_meas: Path = OPTIONS["input_meas"],
    base_output_dir: Path = OPTIONS["output_dir"],
    qc_criteria_path: Path = OPTIONS["qc_criteria"],
    layer: str = OPTIONS["layer"],
    permodule: bool = OPTIONS["permodule"],
    submit: bool = OPTIONS["submit"],
    site: str = OPTIONS["site"],
    nChips: int = OPTIONS["nchips"],
    fit_method: FitMethod = OPTIONS["fit_method"],
    verbosity: LogLevel = OPTIONS["verbosity"],
    lp_enable: bool = OPTIONS["lp_enable"],
):
    log = logging.getLogger(__name__)
    log.setLevel(verbosity.value)

    if submit and site == "":
        log.error(
            bcolors.ERROR
            + "You have supplied the --submit option without specifying --site (testing site). Please specify your testing site if you would like to submit these results"
            + bcolors.ENDC
        )
        return

    log.info("")
    log.info(" =======================================")
    log.info(" \tPerforming SLDO analysis")
    log.info(" =======================================")
    log.info("")

    test_type = os.path.basename(__file__).split(".py")[0]

    # SLDO parameters
    RextA = get_nominal_RextA(layer)
    RextD = get_nominal_RextD(layer)

    allinputs = get_inputs(input_meas, test_type)
    qc_config = get_qc_config(qc_criteria_path, test_type)

    time_start = round(datetime.timestamp(datetime.now()))
    output_dir = base_output_dir.joinpath(test_type).joinpath(f"{time_start}")
    output_dir.mkdir(parents=True, exist_ok=False)

    alloutput = []
    timestamps = []
    for filename in sorted(allinputs):
        log.info("")
        log.info(f" Loading {filename}")
        meas_timestamp = get_time_stamp(filename)
        inputDFs = load_json(filename)

        log.debug(
            f" There are results from {len(inputDFs)} chip(s) stored in this file"
        )
        for inputDF in inputDFs:
            qcframe = inputDF.get_results()
            metadata = qcframe.get_meta_data()

            try:
                kShuntA = (
                    metadata.get("ChipConfigs")
                    .get("RD53B")
                    .get("Parameter")
                    .get("KShuntA")
                )
                log.debug(f" Found kShuntA = {kShuntA} from chip config")
            except Exception:
                log.warning(
                    bcolors.WARNING
                    + " No KShuntA parameter found in chip metadata. Using default KShuntA = 1040"
                    + bcolors.ENDC
                )
                kShuntA = 1040

            try:
                kShuntD = (
                    metadata.get("ChipConfigs")
                    .get("RD53B")
                    .get("Parameter")
                    .get("KShuntD")
                )
                log.debug(f" Found kShuntD = {kShuntD} from chip config")
            except Exception:
                log.warning(
                    bcolors.WARNING
                    + " No KShuntD parameter found in chip metadata. Using default KShuntD = 1040"
                    + bcolors.ENDC
                )
                kShuntD = 1040.0

            try:
                chipname = metadata.get("Name")
                log.debug(f" Found chip name = {chipname} from chip config")
            except Exception:
                log.error(
                    bcolors.ERROR
                    + f" Chip name not found in input from {filename}, skipping."
                    + bcolors.ENDC
                )
                continue

            R_eff = 1.0 / ((kShuntA / RextA) + (kShuntD / RextD)) / nChips

            Vofs = get_nominal_Voffs(layer, lp_enable)

            p = np.poly1d([R_eff, Vofs])
            p1 = np.poly1d([R_eff, 0])

            """"" Check file integrity  """ ""
            checker = JsonChecker(inputDF, test_type)

            try:
                checker.check()
            except BaseException as exc:
                log.exception(exc)
                log.error(
                    bcolors.ERROR
                    + " JsonChecker check not passed, skipping this input."
                    + bcolors.ENDC
                )
                continue
            else:
                log.debug(" JsonChecker check passed!")
                pass

            """""  Calculate quanties   """ ""
            extractor = DataExtractor(inputDF, test_type)
            calculated_data = extractor.calculate()

            passes_qc = True

            # Plot parameters
            Iint_max = (
                max(
                    max(calculated_data["Iref"]["Values"] * 100000),
                    max(calculated_data["IcoreD"]["Values"]),
                    max(calculated_data["IcoreA"]["Values"]),
                    max(calculated_data["IshuntD"]["Values"]),
                    max(calculated_data["IshuntA"]["Values"]),
                    max(calculated_data["IinD"]["Values"]),
                    max(calculated_data["IinA"]["Values"]),
                )
                + 0.5
            )
            I_max = max(calculated_data["SetCurrent"]["Values"]) + 0.5
            I_min = min(calculated_data["SetCurrent"]["Values"]) - 0.5
            V_max = (
                max(
                    max(calculated_data["VrefOVP"]["Values"]),
                    max(calculated_data["Vofs"]["Values"]),
                    max(calculated_data["VDDD"]["Values"]),
                    max(calculated_data["VDDA"]["Values"]),
                    max(calculated_data["VinD"]["Values"]),
                    max(calculated_data["VinA"]["Values"]),
                )
                + 2.0
            )
            T_min = min(0.0, min(calculated_data["Temperature"]["Values"]))
            T_max = max(calculated_data["Temperature"]["Values"]) + 1.0

            # Internal voltages visualization
            fig, ax1 = plt.subplots()
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["VinA"]["Values"],
                marker="o",
                markersize=4,
                label="VinA",
                color="tab:red",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["VinD"]["Values"],
                marker="o",
                markersize=4,
                label="VinD",
                color="tab:red",
                linestyle="--",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["VDDA"]["Values"],
                marker="o",
                markersize=4,
                label="VDDA",
                color="tab:blue",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["VDDD"]["Values"],
                marker="o",
                markersize=4,
                label="VDDD",
                color="tab:blue",
                linestyle="--",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["Vofs"]["Values"],
                marker="o",
                markersize=4,
                label="Vofs",
                color="tab:orange",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["VrefOVP"]["Values"],
                marker="o",
                markersize=4,
                label="VrefOVP",
                color="tab:cyan",
            )

            xp = np.linspace(I_min, I_max, 1000)
            ax1.plot(
                xp,
                p(xp),
                label=f"V = {R_eff:.3f} I + {Vofs:.2f}",
                color="tab:brown",
                linestyle="dotted",
            )
            ax1.set_xlabel("I [A]")
            ax1.set_ylabel("V [V]")
            plt.title(f"VI curve for chip: {chipname}")
            plt.xlim(I_min, I_max)
            ax1.set_ylim(0.0, V_max)
            ax1.legend(loc="upper left", framealpha=0)
            plt.grid()

            ax2 = ax1.twinx()
            ax2.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["Temperature"]["Values"],
                marker="^",
                markersize=4,
                color="tab:green",
                label="Temperature (NTC)",
                linestyle="-.",
            )
            ax2.set_ylabel("T [C]")
            ax2.set_ylim(T_min, T_max)
            ax2.legend(loc="upper right", framealpha=0)

            plt.tight_layout()
            outfile = output_dir.joinpath(f"{chipname}_VI.png")
            log.info(f" Saving {outfile}")
            plt.savefig(outfile)
            plt.clf()

            ax1.cla()
            ax2.cla()
            f = plt.figure()
            f.clear()
            plt.close(f)

            # Internal currents visualization
            fig, ax1 = plt.subplots()
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["IinA"]["Values"],
                marker="o",
                markersize=4,
                label="IinA",
                color="tab:red",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["IinD"]["Values"],
                marker="o",
                markersize=4,
                label="IinD",
                color="tab:red",
                linestyle="--",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["IshuntA"]["Values"],
                marker="o",
                markersize=4,
                label="IshuntA",
                color="tab:blue",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["IshuntD"]["Values"],
                marker="o",
                markersize=4,
                label="IshuntD",
                color="tab:blue",
                linestyle="--",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["IcoreA"]["Values"],
                marker="o",
                markersize=4,
                label="IcoreA",
                color="tab:orange",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["IcoreD"]["Values"],
                marker="o",
                markersize=4,
                label="IcoreD",
                color="tab:orange",
                linestyle="--",
            )
            ax1.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["Iref"]["Values"] * 100000,
                marker="o",
                markersize=4,
                label="Iref*100k",
                color="tab:cyan",
            )
            ax1.set_xlabel("I [A]")
            ax1.set_ylabel("I [A]")
            plt.title(f"Currents for chip: {chipname}")

            plt.xlim(I_min, I_max)
            plt.ylim(0.0, Iint_max)
            ax1.legend(loc="upper left", framealpha=0)
            plt.grid()

            ax2 = ax1.twinx()
            ax2.plot(
                calculated_data["SetCurrent"]["Values"],
                calculated_data["Temperature"]["Values"],
                marker="^",
                markersize=4,
                color="tab:green",
                label="Temperature (NTC)",
                linestyle="-.",
            )
            ax2.set_ylabel("T [C]")
            ax2.set_ylim(T_min, T_max)
            ax2.legend(loc="upper right", framealpha=0)

            plt.tight_layout()
            outfile = output_dir.joinpath(f"{chipname}_II.png")
            log.info(f" Saving {outfile}")
            plt.savefig(outfile)
            plt.clf()

            ax1.cla()
            ax2.cla()
            f = plt.figure()
            f.clear()
            plt.close(f)

            # SLDO fit
            VinAvg = (
                calculated_data["VinA"]["Values"] + calculated_data["VinD"]["Values"]
            ) / 2.0
            if fit_method.value == "root":
                slope, offset = linear_fit(
                    calculated_data["SetCurrent"]["Values"], VinAvg
                )
            else:
                slope, offset = linear_fit_np(
                    calculated_data["SetCurrent"]["Values"], VinAvg
                )

            # Residual analysis
            residual_VinA = (
                p1(calculated_data["SetCurrent"]["Values"])
                - (
                    calculated_data["VinA"]["Values"]
                    - calculated_data["Vofs"]["Values"]
                )
            ) * 1000
            residual_VinD = (
                p1(calculated_data["SetCurrent"]["Values"])
                - (
                    calculated_data["VinD"]["Values"]
                    - calculated_data["Vofs"]["Values"]
                )
            ) * 1000
            residual_VinA_nomVofs = (
                p(calculated_data["SetCurrent"]["Values"])
                - calculated_data["VinA"]["Values"]
            ) * 1000
            residual_VinD_nomVofs = (
                p(calculated_data["SetCurrent"]["Values"])
                - calculated_data["VinD"]["Values"]
            ) * 1000
            residual_Vin = p1(calculated_data["SetCurrent"]["Values"]) - (
                VinAvg - calculated_data["Vofs"]["Values"]
            )
            residual_Vofs = (Vofs - calculated_data["Vofs"]["Values"]) * 1000
            res_max = (
                max(
                    max(residual_VinA_nomVofs),
                    max(residual_VinD_nomVofs),
                    max(residual_VinA),
                    max(residual_VinD),
                    max(residual_Vofs),
                )
                + 20
            )
            res_min = (
                min(
                    min(residual_VinA_nomVofs),
                    min(residual_VinD_nomVofs),
                    min(residual_VinA),
                    min(residual_VinD),
                    min(residual_Vofs),
                )
                - 10
            )

            plt.plot(
                calculated_data["SetCurrent"]["Values"],
                residual_VinA_nomVofs,
                marker="o",
                markersize=4,
                label=f"{R_eff:.3f}I+{Vofs:.2f}-VinA",
                color="tab:red",
            )
            plt.plot(
                calculated_data["SetCurrent"]["Values"],
                residual_VinD_nomVofs,
                marker="o",
                markersize=4,
                label=f"{R_eff:.3f}I+{Vofs:.2f}-VinD",
                color="tab:red",
                linestyle="--",
            )
            plt.plot(
                calculated_data["SetCurrent"]["Values"],
                residual_VinA,
                marker="o",
                markersize=4,
                label=f"{R_eff:.3f}I+Vofs-VinA",
                color="tab:blue",
            )
            plt.plot(
                calculated_data["SetCurrent"]["Values"],
                residual_VinD,
                marker="o",
                markersize=4,
                label=f"{R_eff:.3f}I+Vofs-VinD",
                color="tab:blue",
                linestyle="--",
            )
            plt.plot(
                calculated_data["SetCurrent"]["Values"],
                residual_Vofs,
                marker="o",
                markersize=4,
                label=f"{Vofs}-Vofs",
                color="tab:orange",
            )
            plt.xlabel("I [A]")
            plt.ylabel("V [mV]")
            plt.title(f"VI curve for chip: {chipname}")
            plt.xlim(I_min, I_max)
            plt.ylim(res_min, res_max)
            plt.legend(loc="upper right", framealpha=0)
            plt.grid()
            plt.tight_layout()
            outfile = output_dir.joinpath(f"{chipname}_VIresidual.png")
            log.info(f" Saving {outfile}")
            plt.savefig(outfile)
            plt.clf()

            f = plt.figure()
            f.clear()
            plt.close(f)

            # Find point measured closest to nominal input current
            sldo_nom_input_current = get_nominal_current(layer)
            log.debug(f" Calculated nominal current to be: {sldo_nom_input_current}")
            idx = (
                np.abs(calculated_data["SetCurrent"]["Values"] - sldo_nom_input_current)
            ).argmin()
            log.debug(
                f' Closest current measured to nominal is: {calculated_data["SetCurrent"]["Values"][idx]}'
            )

            # Load values to dictionary for QC analysis
            results = {}
            results.update({"SLDOLinearity": max(residual_Vin)})
            results.update(
                {
                    "VInA_VInD": max(
                        abs(
                            calculated_data["VinA"]["Values"]
                            - calculated_data["VinD"]["Values"]
                        )
                    )
                }
            )
            results.update({"SLDO_VDDA": calculated_data["VDDA"]["Values"][idx]})
            results.update({"SLDO_VDDD": calculated_data["VDDD"]["Values"][idx]})
            results.update({"SLDO_VINA": calculated_data["VinA"]["Values"][idx]})
            results.update({"SLDO_VIND": calculated_data["VinD"]["Values"][idx]})
            results.update({"SLDO_VOFFS": calculated_data["Vofs"]["Values"][idx]})
            results.update({"SLDO_IINA": calculated_data["IinA"]["Values"][idx]})
            results.update({"SLDO_IIND": calculated_data["IinD"]["Values"][idx]})
            results.update({"SLDO_IREF": calculated_data["Iref"]["Values"][idx] * 1e6})
            results.update({"SLDO_ISHUNTA": calculated_data["IshuntA"]["Values"][idx]})
            results.update({"SLDO_ISHUNTD": calculated_data["IshuntD"]["Values"][idx]})

            # Perform QC analysis
            passes_qc = perform_qc_analysis(
                test_type, qc_config, layer, results, verbosity.value
            )
            if passes_qc == -1:
                log.error(
                    bcolors.ERROR
                    + f" QC analysis for {chipname} was NOT successful. Please fix and re-run. Continuing to next chip.."
                    + bcolors.ENDC
                )
                continue
            log.info("")
            if passes_qc:
                log.info(
                    f" Chip {chipname} passes QC? "
                    + bcolors.OKGREEN
                    + f"{passes_qc}"
                    + bcolors.ENDC
                )
            else:
                log.info(
                    f" Chip {chipname} passes QC? "
                    + bcolors.BADRED
                    + f"{passes_qc}"
                    + bcolors.ENDC
                )
            log.info("")

            """"" Output a json file """ ""
            outputDF = outputDataFrame()
            outputDF.set_test_type(test_type)
            data = qcDataFrame()
            data._meta_data.update(metadata)
            data.add_property(
                "ANALYSIS_VERSION",
                pkg_resources.get_distribution("module-qc-analysis-tools").version,
            )
            # Load values to store in output file
            analog_overhead = calculated_data["IshuntA"]["Values"][idx] / (
                calculated_data["IinA"]["Values"][idx]
                - calculated_data["IshuntA"]["Values"][idx]
            )
            digital_overhead = calculated_data["IshuntD"]["Values"][idx] / (
                calculated_data["IinD"]["Values"][idx]
                - calculated_data["IshuntD"]["Values"][idx]
            )
            data.add_parameter("SLDO_VI_SLOPE", round(slope, 4))
            data.add_parameter("SLDO_VI_OFFSET", round(offset, 4))
            data.add_parameter(
                "SLDO_NOM_INPUT_CURRENT", round(sldo_nom_input_current, 4)
            )
            data.add_parameter(
                "SLDO_VDDA", round(calculated_data["VDDA"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_VDDD", round(calculated_data["VDDD"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_VINA", round(calculated_data["VinA"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_VIND", round(calculated_data["VinD"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_VOFFS", round(calculated_data["Vofs"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_IINA", round(calculated_data["IinA"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_IIND", round(calculated_data["IinD"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_IREF", round(calculated_data["Iref"]["Values"][idx] * 1e6, 4)
            )
            data.add_parameter(
                "SLDO_ISHUNTA", round(calculated_data["IshuntA"]["Values"][idx], 4)
            )
            data.add_parameter(
                "SLDO_ISHUNTD", round(calculated_data["IshuntD"]["Values"][idx], 4)
            )
            data.add_parameter("SLDO_ANALOG_OVERHEAD", round(analog_overhead, 4))
            data.add_parameter("SLDO_DIGITAL_OVERHEAD", round(digital_overhead, 4))

            outputDF.set_results(data)
            outputDF.set_pass_flag(passes_qc)
            if submit:
                submit_results(
                    outputDF.to_dict(True),
                    time_start,
                    site,
                    output_dir.joinpath("submit.txt"),
                )
            if permodule:
                alloutput += [outputDF.to_dict(True)]
                timestamps += [meas_timestamp]
            else:
                outfile = output_dir.joinpath(f"{chipname}.json")
                log.info(f" Saving output of analysis to: {outfile}")
                save_dict_list(outfile, [outputDF.to_dict(True)])
    if permodule:
        # Only store results from same timestamp into same file
        dfs = np.array(alloutput)
        tss = np.array(timestamps)
        for x in np.unique(tss):
            outfile = output_dir.joinpath("module.json")
            log.info(f" Saving output of analysis to: {outfile}")
            save_dict_list(
                outfile,
                dfs[tss == x].tolist(),
            )


if __name__ == "__main__":
    typer.run(main)



from __future__ import annotations

import json
import logging
import sys
from pathlib import Path
from typing import Any

import numpy as np

if sys.version_info >= (3, 9):
    from importlib import resources
else:
    import importlib_resources as resources

log = logging.getLogger(__name__)
log.setLevel("INFO")


class JsonFileCheckFailure(Exception):
    pass


class LengthMismatchError(Exception):
    pass


class NegativeMeasurementError(Exception):
    pass


class bcolors:
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    ERROR = "\033[91m"
    BADRED = "\033[91m"
    ENDC = "\033[0m"


def getVmuxMap():
    Vmux_map = {
        0: "GADC",
        1: "ImuxPad",
        2: "Vntc",
        3: "VcalDac",
        4: "VDDAcapmeas",
        5: "VPolSensTop",
        6: "VPolSensBottom",
        7: "VcalHi",
        8: "VcalMed",
        9: "VDiffVTH2",
        10: "VDiffVTH1Main",
        11: "VDiffVTH1Left",
        12: "VDiffVTH1Right",
        13: "VRadSensAna",
        14: "VMonSensAna",
        15: "VRadSensDig",
        16: "VMonSensDig",
        17: "VRadSensCenter",
        18: "VMonSensAcb",
        19: "AnaGND19",
        20: "AnaGND20",
        21: "AnaGND21",
        22: "AnaGND22",
        23: "AnaGND23",
        24: "AnaGND24",
        25: "AnaGND25",
        26: "AnaGND26",
        27: "AnaGND27",
        28: "AnaGND28",
        29: "AnaGND29",
        30: "AnaGND30",
        31: "VrefCore",
        32: "VrefOVP",
        33: "VinA",
        34: "VDDA",
        35: "VrefA",
        36: "Vofs",
        37: "VinD",
        38: "VDDD",
        39: "VrefD",
    }

    return Vmux_map


def getImuxMap():
    Imux_map = {
        0: "Iref",
        1: "CdrVcoMainBias",
        2: "CdrVcoBuffBias",
        3: "ICdrCP",
        4: "ICdrFD",
        5: "CdrBuffBias",
        6: "CMLDrivTap2Bias",
        7: "CMLDrivTap1Bias",
        8: "CMLDrivMainBias",
        9: "Intc",
        10: "CapMeas",
        11: "CapMeasPar",
        12: "IDiffPreMain",
        13: "IDiffPreampComp",
        14: "IDiffComp",
        15: "IDiffVth2",
        16: "IDiffVth1Main",
        17: "IDiffLcc",
        18: "IDiffFB",
        19: "IDiffPreampLeft",
        20: "IDiffVth1Left",
        21: "IDiffPreampRight",
        22: "IDiffPreampTopLeft",
        23: "IDiffVth1Right",
        24: "IDiffPreampTop",
        25: "IDiffPreampTopRight",
        26: "NotUsed26",
        27: "NotUsed27",
        28: "IinA",
        29: "IshuntA",
        30: "IinD",
        31: "IshuntD",
    }
    return Imux_map


class JsonChecker:
    def __init__(self, inputDF, test_type_from_script) -> None:
        # init all available qc types
        self.qc_type = test_type_from_script
        self.inputdataframe = inputDF
        self.qcdataframe = inputDF.get_results()

        self.dict_map = {
            "VCAL_CALIBRATION": {
                "VCAL_HIGH": {
                    "required_keys": ["DACs_input", "Vmux30", "Vmux7"],
                    "InjVcalRange": 1,
                    "MonitorV": 7,
                },
                "VCAL_HIGH_SMALL_RANGE": {
                    "required_keys": ["DACs_input", "Vmux30", "Vmux7"],
                    "InjVcalRange": 0,
                    "MonitorV": 7,
                },
                "VCAL_MED": {
                    "required_keys": ["DACs_input", "Vmux30", "Vmux8"],
                    "InjVcalRange": 1,
                    "MonitorV": 8,
                },
                "VCAL_MED_SMALL_RANGE": {
                    "required_keys": ["DACs_input", "Vmux30", "Vmux8"],
                    "InjVcalRange": 0,
                    "MonitorV": 8,
                },
            },
            "ADC_CALIBRATION": {
                "required_keys": ["DACs_input", "ADC_Vmux8", "Vmux30", "Vmux8"],
                "InjVcalRange": 1,
            },
            "SLDO": {
                "required_keys": [
                    "Temperature",
                    "Current",
                    "Vmux30",
                    "Vmux33",
                    "Vmux37",
                    "Vmux34",
                    "Vmux38",
                    "Vmux36",
                    "Vmux32",
                    "Imux0",
                    "Imux28",
                    "Imux29",
                    "Imux30",
                    "Imux31",
                    "Imux63",
                ]
            },
            "INJECTION_CAPACITANCE": {
                "required_keys": ["Vmux4", "Vmux30", "Imux10", "Imux11", "Imux63"]
            },
            "ANALOG_READBACK": {
                "AR_VMEAS": {
                    "required_keys": [
                        "Vmux30",
                        "Vmux33",
                        "Vmux34",
                        "Vmux35",
                        "Vmux36",
                        "Vmux37",
                        "Vmux38",
                        "Imux28",
                        "Imux29",
                        "Imux30",
                        "Imux31",
                        "Imux63",
                    ],
                },
                "AR_TEMP": {
                    "required_keys": [
                        "Vmux14",
                        "Vmux16",
                        "Vmux18",
                        "Vmux2",
                        "Vmux30",
                        "Imux9",
                        "Imux63",
                        "TExtExtNTC",
                    ],
                },
                "AR_VDD": {
                    "required_keys": [
                        "Vmux34",
                        "Vmux38",
                        "Vmux30",
                    ],
                },
            },
        }

    def check_testtype(self):
        if self.inputdataframe._subtestType != "":
            self.sub_test_type = self.inputdataframe._subtestType
            required_testtype = self.qc_type
            self.required_keywords = self.dict_map[self.qc_type][self.sub_test_type][
                "required_keys"
            ]
        else:
            required_testtype = self.qc_type
            self.required_keywords = self.dict_map[self.qc_type]["required_keys"]

        testtype_from_file = self.inputdataframe._testType
        if not required_testtype == testtype_from_file:
            log.error(
                bcolors.ERROR
                + f"Required testtype of the file '{required_testtype}' is not matched to '{testtype_from_file}' from the file! "
                + bcolors.ENDC
            )
            raise KeyError()

    def check_keywords_exist(self) -> None:
        for required_keyword in self.required_keywords:
            if required_keyword not in self.qcdataframe._data.keys():
                log.error(
                    bcolors.ERROR + f"{required_keyword} not found! " + bcolors.ENDC
                )
                raise KeyError()

    def check_keywords_length(self) -> None:
        """Check whether the length of measurements are same."""
        first_keyword = True
        for required_keyword in self.required_keywords:
            if first_keyword:
                len_x = len(self.qcdataframe[required_keyword])
                self.var_x = required_keyword
                first_keyword = False
            if self.qcdataframe.get_x(required_keyword) is True:
                len_x = len(self.qcdataframe[required_keyword])
                self.var_x = required_keyword

        for required_keyword in self.required_keywords:
            if len_x != len(self.qcdataframe[required_keyword]):
                log.error(
                    bcolors.ERROR
                    + f"The length of {required_keyword} is not equal to {self.var_x}!"
                    + bcolors.ENDC
                )
                raise LengthMismatchError()

    def check_positive_values(self) -> None:
        """Check whether the contents of measurements are valid, i.e. all positive."""
        for required_keyword in self.required_keywords:
            if ((self.qcdataframe[required_keyword]) < 0).sum() > 0:
                log.error(
                    bcolors.ERROR
                    + f"Negative measurements observed in {required_keyword}"
                    + bcolors.ENDC
                )
                raise NegativeMeasurementError()

    def check_parameters_overwritten(self) -> None:
        """Check whether the parameters are overwritten correctly."""
        if self.inputdataframe._subtestType != "":
            parameters_to_check = self.dict_map[self.qc_type][self.sub_test_type].copy()
        else:
            parameters_to_check = self.dict_map[self.qc_type].copy()

        parameters_to_check.pop("required_keys")

        for k, v in parameters_to_check.items():
            value_from_file = lookup(self.qcdataframe._meta_data, k)
            if value_from_file is None:
                log.error(
                    bcolors.ERROR
                    + f"Measurement output file Corrupted! Please check this key: {k}"
                    + bcolors.ENDC
                )
                raise KeyError()
            if v != value_from_file:
                log.error(
                    bcolors.ERROR
                    + f"Values Mismathed for this key {k}: {value_from_file} from file, but required {v} "
                    + bcolors.ENDC
                )
                raise KeyError()

    def check(self) -> None:
        self.check_testtype()
        self.check_keywords_exist()
        self.check_keywords_length()
        self.check_positive_values()
        self.check_parameters_overwritten()


class DataExtractor(JsonChecker):
    def __init__(self, inputDF, test_type_from_script) -> None:
        super().__init__(inputDF, test_type_from_script)
        qcframe = inputDF.get_results()
        self.df = qcframe._data
        self.rImux = qcframe._meta_data["ChipConfigs"]["RD53B"]["Parameter"].get(
            "R_Imux", 10000.0
        )
        self.kIinA = qcframe._meta_data["ChipConfigs"]["RD53B"]["Parameter"].get(
            "KSenseInA", 21000.0
        )
        self.kIinD = qcframe._meta_data["ChipConfigs"]["RD53B"]["Parameter"].get(
            "KSenseInD", 21000.0
        )
        self.kIshuntA = qcframe._meta_data["ChipConfigs"]["RD53B"]["Parameter"].get(
            "KSenseShuntA", 26000.0
        )
        self.kIshuntD = qcframe._meta_data["ChipConfigs"]["RD53B"]["Parameter"].get(
            "KSenseShuntD", 26000.0
        )
        self.VmuxGrd = "Vmux30"
        self.ImuxGrd = "Imux63"

    def getQuantity(self, key):
        return getattr(self, key, lambda: False)()

    def Vmux0(self):
        return {
            getVmuxMap()[0]: {
                "X": self.df["Vmux0"]["X"],
                "Unit": self.df["Vmux0"]["Unit"],
                "Values": (
                    self.df["Vmux0"]["Values"] - self.df[self.VmuxGrd]["Values"]
                ),
            }
        }

    def Vmux2(self):
        return {
            getVmuxMap()[2]: {
                "X": self.df["Vmux2"]["X"],
                "Unit": self.df["Vmux2"]["Unit"],
                "Values": (
                    self.df["Vmux2"]["Values"] - self.df[self.VmuxGrd]["Values"]
                ),
            }
        }

    def Vmux3(self):
        return {
            getVmuxMap()[3]: {
                "X": self.df["Vmux3"]["X"],
                "Unit": self.df["Vmux3"]["Unit"],
                "Values": (self.df["Vmux3"]["Values"] - self.df[self.VmuxGrd]["Values"])
                * 2,
            }
        }

    def Vmux4(self):
        return {
            getVmuxMap()[4]: {
                "X": self.df["Vmux4"]["X"],
                "Unit": self.df["Vmux4"]["Unit"],
                "Values": (self.df["Vmux4"]["Values"] - self.df[self.VmuxGrd]["Values"])
                * 2,
            }
        }

    def Vmux5(self):
        return {
            getVmuxMap()[5]: {
                "X": self.df["Vmux5"]["X"],
                "Unit": self.df["Vmux5"]["Unit"],
                "Values": self.df["Vmux5"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux6(self):
        return {
            getVmuxMap()[6]: {
                "X": self.df["Vmux6"]["X"],
                "Unit": self.df["Vmux6"]["Unit"],
                "Values": self.df["Vmux6"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux7(self):
        return {
            getVmuxMap()[7]: {
                "X": self.df["Vmux7"]["X"],
                "Unit": self.df["Vmux7"]["Unit"],
                "Values": self.df["Vmux7"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux8(self):
        return {
            getVmuxMap()[8]: {
                "X": self.df["Vmux8"]["X"],
                "Unit": self.df["Vmux8"]["Unit"],
                "Values": self.df["Vmux8"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux9(self):
        return {
            getVmuxMap()[9]: {
                "X": self.df["Vmux9"]["X"],
                "Unit": self.df["Vmux9"]["Unit"],
                "Values": self.df["Vmux9"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux10(self):
        return {
            getVmuxMap()[10]: {
                "X": self.df["Vmux10"]["X"],
                "Unit": self.df["Vmux10"]["Unit"],
                "Values": self.df["Vmux10"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux11(self):
        return {
            getVmuxMap()[11]: {
                "X": self.df["Vmux11"]["X"],
                "Unit": self.df["Vmux11"]["Unit"],
                "Values": self.df["Vmux11"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux12(self):
        return {
            getVmuxMap()[12]: {
                "X": self.df["Vmux12"]["X"],
                "Unit": self.df["Vmux12"]["Unit"],
                "Values": self.df["Vmux12"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux13(self):
        return {
            getVmuxMap()[13]: {
                "X": self.df["Vmux13"]["X"],
                "Unit": self.df["Vmux13"]["Unit"],
                "Values": self.df["Vmux13"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux14(self):
        return {
            getVmuxMap()[14]: {
                "X": self.df["Vmux14"]["X"],
                "Unit": self.df["Vmux14"]["Unit"],
                "Values": self.df["Vmux14"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux15(self):
        return {
            getVmuxMap()[15]: {
                "X": self.df["Vmux15"]["X"],
                "Unit": self.df["Vmux15"]["Unit"],
                "Values": self.df["Vmux15"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux16(self):
        return {
            getVmuxMap()[16]: {
                "X": self.df["Vmux16"]["X"],
                "Unit": self.df["Vmux16"]["Unit"],
                "Values": self.df["Vmux16"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux17(self):
        return {
            getVmuxMap()[17]: {
                "X": self.df["Vmux17"]["X"],
                "Unit": self.df["Vmux17"]["Unit"],
                "Values": self.df["Vmux17"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux18(self):
        return {
            getVmuxMap()[18]: {
                "X": self.df["Vmux18"]["X"],
                "Unit": self.df["Vmux18"]["Unit"],
                "Values": self.df["Vmux18"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux30(self):
        return {
            getVmuxMap()[30]: {
                "X": self.df["Vmux30"]["X"],
                "Unit": self.df["Vmux30"]["Unit"],
                "Values": self.df["Vmux30"]["Values"],
            }
        }

    def Vmux31(self):
        return {
            getVmuxMap()[31]: {
                "X": self.df["Vmux31"]["X"],
                "Unit": self.df["Vmux31"]["Unit"],
                "Values": self.df["Vmux31"]["Values"] - self.df[self.VmuxGrd]["Values"],
            }
        }

    def Vmux32(self):
        return {
            getVmuxMap()[32]: {
                "X": self.df["Vmux32"]["X"],
                "Unit": self.df["Vmux32"]["Unit"],
                "Values": (
                    self.df["Vmux32"]["Values"] - self.df[self.VmuxGrd]["Values"]
                )
                * 3.33,
            }
        }

    def Vmux33(self):
        return {
            getVmuxMap()[33]: {
                "X": self.df["Vmux33"]["X"],
                "Unit": self.df["Vmux33"]["Unit"],
                "Values": (
                    self.df["Vmux33"]["Values"] - self.df[self.VmuxGrd]["Values"]
                )
                * 4,
            }
        }

    def Vmux34(self):
        return {
            getVmuxMap()[34]: {
                "X": self.df["Vmux34"]["X"],
                "Unit": self.df["Vmux34"]["Unit"],
                "Values": (
                    self.df["Vmux34"]["Values"] - self.df[self.VmuxGrd]["Values"]
                )
                * 2,
            }
        }

    def Vmux35(self):
        return {
            getVmuxMap()[35]: {
                "X": self.df["Vmux35"]["X"],
                "Unit": self.df["Vmux35"]["Unit"],
                "Values": (
                    self.df["Vmux35"]["Values"] - self.df[self.VmuxGrd]["Values"]
                ),
            }
        }

    def Vmux36(self):
        return {
            getVmuxMap()[36]: {
                "X": self.df["Vmux36"]["X"],
                "Unit": self.df["Vmux36"]["Unit"],
                "Values": (
                    self.df["Vmux36"]["Values"] - self.df[self.VmuxGrd]["Values"]
                )
                * 4,
            }
        }

    def Vmux37(self):
        return {
            getVmuxMap()[37]: {
                "X": self.df["Vmux37"]["X"],
                "Unit": self.df["Vmux37"]["Unit"],
                "Values": (
                    self.df["Vmux37"]["Values"] - self.df[self.VmuxGrd]["Values"]
                )
                * 4,
            }
        }

    def Vmux38(self):
        return {
            getVmuxMap()[38]: {
                "X": self.df["Vmux38"]["X"],
                "Unit": self.df["Vmux38"]["Unit"],
                "Values": (
                    self.df["Vmux38"]["Values"] - self.df[self.VmuxGrd]["Values"]
                )
                * 2,
            }
        }

    def Vmux39(self):
        return {
            getVmuxMap()[39]: {
                "X": self.df["Vmux39"]["X"],
                "Unit": self.df["Vmux39"]["Unit"],
                "Values": (
                    self.df["Vmux39"]["Values"] - self.df[self.VmuxGrd]["Values"]
                ),
            }
        }

    def Imux0(self):
        return {
            getImuxMap()[0]: {
                "X": self.df["Imux0"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux0"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux1(self):
        return {
            getImuxMap()[1]: {
                "X": self.df["Imux1"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux1"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux2(self):
        return {
            getImuxMap()[2]: {
                "X": self.df["Imux2"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux2"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux3(self):
        return {
            getImuxMap()[3]: {
                "X": self.df["Imux3"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux3"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux4(self):
        return {
            getImuxMap()[4]: {
                "X": self.df["Imux4"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux4"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux5(self):
        return {
            getImuxMap()[5]: {
                "X": self.df["Imux5"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux5"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux6(self):
        return {
            getImuxMap()[6]: {
                "X": self.df["Imux6"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux6"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux7(self):
        return {
            getImuxMap()[7]: {
                "X": self.df["Imux7"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux7"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux8(self):
        return {
            getImuxMap()[8]: {
                "X": self.df["Imux8"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux8"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux9(self):
        return {
            getImuxMap()[9]: {
                "X": self.df["Imux9"]["X"],
                "Unit": "A",
                "Values": (self.df["Imux9"]["Values"] - self.df[self.ImuxGrd]["Values"])
                / self.rImux,
            }
        }

    def Imux10(self):
        return {
            getImuxMap()[10]: {
                "X": self.df["Imux10"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux10"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux11(self):
        return {
            getImuxMap()[11]: {
                "X": self.df["Imux11"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux11"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux12(self):
        return {
            getImuxMap()[12]: {
                "X": self.df["Imux12"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux12"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux13(self):
        return {
            getImuxMap()[13]: {
                "X": self.df["Imux13"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux13"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux14(self):
        return {
            getImuxMap()[14]: {
                "X": self.df["Imux14"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux14"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux15(self):
        return {
            getImuxMap()[15]: {
                "X": self.df["Imux15"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux15"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux16(self):
        return {
            getImuxMap()[16]: {
                "X": self.df["Imux16"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux16"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux17(self):
        return {
            getImuxMap()[17]: {
                "X": self.df["Imux17"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux17"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux18(self):
        return {
            getImuxMap()[18]: {
                "X": self.df["Imux18"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux18"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux19(self):
        return {
            getImuxMap()[19]: {
                "X": self.df["Imux15"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux15"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux20(self):
        return {
            getImuxMap()[20]: {
                "X": self.df["Imux20"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux20"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux21(self):
        return {
            getImuxMap()[21]: {
                "X": self.df["Imux21"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux21"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux22(self):
        return {
            getImuxMap()[22]: {
                "X": self.df["Imux22"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux22"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux23(self):
        return {
            getImuxMap()[23]: {
                "X": self.df["Imux23"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux23"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux24(self):
        return {
            getImuxMap()[24]: {
                "X": self.df["Imux24"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux24"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux25(self):
        return {
            getImuxMap()[25]: {
                "X": self.df["Imux25"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux25"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux,
            }
        }

    def Imux28(self):
        return {
            getImuxMap()[28]: {
                "X": self.df["Imux28"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux28"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux
                * self.kIinA,
            }
        }

    def Imux29(self):
        return {
            getImuxMap()[29]: {
                "X": self.df["Imux29"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux29"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux
                * self.kIinD,
            }
        }

    def Imux30(self):
        return {
            getImuxMap()[30]: {
                "X": self.df["Imux30"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux30"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux
                * self.kIshuntA,
            }
        }

    def Imux31(self):
        return {
            getImuxMap()[31]: {
                "X": self.df["Imux31"]["X"],
                "Unit": "A",
                "Values": (
                    self.df["Imux31"]["Values"] - self.df[self.ImuxGrd]["Values"]
                )
                / self.rImux
                * self.kIshuntD,
            }
        }

    def IcoreA(self):
        return {
            "IcoreA": {
                "X": False,
                "Unit": "A",
                "Values": self.df["IinA"]["Values"] - self.df["IshuntA"]["Values"],
            }
        }

    def IcoreD(self):
        return {
            "IcoreD": {
                "X": False,
                "Unit": "A",
                "Values": self.df["IinD"]["Values"] - self.df["IshuntD"]["Values"],
            }
        }

    def Iin(self):
        return {
            "Iin": {
                "X": False,
                "Unit": "A",
                "Values": self.df["IinA"]["Values"] + self.df["IinD"]["Values"],
            }
        }

    def calculate(self):

        # Convert values from list to numpy array
        for key, value in self.df.items():
            self.df.update(
                {
                    key: {
                        "X": value["X"],
                        "Unit": value["Unit"],
                        "Values": np.array(value["Values"]),
                    }
                }
            )

        df_keys = self.df.copy().keys()

        # Calculate basic quantities
        for key in df_keys:
            newQuantity = self.getQuantity(key)
            if newQuantity:
                self.df.update(newQuantity)

        # Calculated complex quantities
        if self.df.get("IinA") and self.df.get("IshuntA"):
            self.df.update(self.getQuantity("IcoreA"))
        if self.df.get("IinD") and self.df.get("IshuntD"):
            self.df.update(self.getQuantity("IcoreD"))
        if self.df.get("IinA") and self.df.get("IinD"):
            self.df.update(self.getQuantity("Iin"))

        return self.df


def linear_fit(x, y):
    try:
        import ROOT
    except ImportError:
        log.error(bcolors.ERROR + "No PyROOT module found! " + bcolors.ENDC)

    from array import array

    """Linear fit with PyROOT

    Args:
        x (array): the x values, input DACs for VCal calibration
        y (array): the y values, measured Vmux[int] for VCal calibration

    Returns:
        result.Parameter(1) (float): the slope of the fitted line
        result.Parameter(0) (float): the offset of the fitted line
    """
    gr = ROOT.TGraph(len(x), array("d", x), array("d", y))

    f = ROOT.TF1("f", "[1] * x + [0]")
    result = gr.Fit(f, "S")

    return result.Parameter(1), result.Parameter(0)


def linear_fit_np(x, y):
    """Linear fit with NumPy. numpy.linalg.lstsq() is called.

    Args:
        x (array): the x values, input DACs for VCal calibration
        y (array): the y values, measured Vmux[int] for VCal calibration

    Returns:
        p1 (float): the slope of the fitted line
        p0 (float): the offset of the fitted line
    """
    A = np.vstack([x, np.ones(len(x))]).T
    p1, p0 = np.linalg.lstsq(A, y, rcond=None)[0]
    return p1, p0


def lookup(input_dict, search_key):
    # recursion to look up the desired key in a dict and record the path
    for k, v in input_dict.items():
        if k == search_key:
            return v
        elif isinstance(v, dict):
            _v = lookup(v, search_key)
            if _v is not None:
                return _v


def get_inputs(input_meas: Path, test_type: str, isConfig=False) -> list[Path]:
    # Figure out if input if single file or directory
    allinputs = []
    if input_meas.is_file():
        allinputs = [input_meas]
    elif input_meas.is_dir():
        if (test_type not in str(input_meas)) and (not isConfig):
            log.error(
                bcolors.ERROR
                + f"The input path {input_meas} does not include the test type {test_type}. Please check the input path."
                + bcolors.ENDC
            )
            raise FileNotFoundError()
        allinputs = sorted(input_meas.glob("*.json"))
        if not allinputs:
            log.error(
                bcolors.ERROR
                + f"No input json files in `{input_meas}` are found! Please check the input path."
                + bcolors.ENDC
            )
            raise FileNotFoundError()
    else:
        log.error(
            bcolors.ERROR
            + "Input is not recognized as single file or path to directory containing files. Please check the input."
            + bcolors.ENDC
        )
        raise FileNotFoundError()
    return allinputs


def get_time_stamp(filename):
    # Read timestamp from measurement output file name
    # Assumes that filename ends with pattern: "/{timestamp}/{chipname}.json"
    return filename.parent.name


def get_qc_config(
    qc_criteria_path: Path, test_type: str | None = None
) -> dict[str, Any]:
    try:
        with resources.as_file(qc_criteria_path) as path:
            qc_config = json.loads(path.read_text())
    except Exception as e:
        log.debug(e)
        log.error(
            bcolors.ERROR
            + f" QC criteria json file is ill-formated ({qc_criteria_path}) - please fix! Exiting."
            + bcolors.ENDC
        )
        raise RuntimeError() from e

    if test_type and not qc_config.get(test_type):
        log.error(
            bcolors.ERROR
            + f" QC criteria for {test_type} not found in {qc_criteria_path} - please fix! Exiting."
            + bcolors.ENDC
        )
        raise FileNotFoundError()

    return qc_config



from __future__ import annotations

from enum import Enum
from pathlib import Path

import typer

from module_qc_analysis_tools import data

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


class FitMethod(str, Enum):
    root = "root"
    numpy = "numpy"


class LogLevel(str, Enum):
    debug = "DEBUG"
    info = "INFO"
    warning = "WARNING"
    error = "ERROR"


OPTIONS = {}

OPTIONS["input_meas"]: Path = typer.Option(
    ...,
    "-i",
    "--input-meas",
    help="path to the input measurement file or directory containing input measurement files.",
    exists=True,
    file_okay=True,
    readable=True,
    resolve_path=True,
)
OPTIONS["output_dir"]: Path = typer.Option(
    "outputs",
    "-o",
    "--output-dir",
    help="output directory",
    exists=False,
    writable=True,
)
OPTIONS["qc_criteria"]: Path = typer.Option(
    str(data / "analysis_cuts.json"),
    "-q",
    "--qc-criteria",
    help="path to json file with QC selection criteria (default: $(module-qc-analysis-tools --prefix)/analysis_cuts.json)",
    exists=True,
    file_okay=True,
    readable=True,
    resolve_path=True,
)
OPTIONS["layer"]: str = typer.Option(
    "Unknown",
    "-l",
    "--layer",
    help="Layer of module, used for applying correct QC criteria settings. Options: L0, L1, L2 (default)",
)
OPTIONS["permodule"]: bool = typer.Option(
    False, help="Store results in one file per module (default: one file per chip)"
)
OPTIONS["fit_method"]: FitMethod = typer.Option(
    FitMethod.numpy, "-f", "--fit-method", help="fitting method"
)
OPTIONS["nchips"]: int = typer.Option(
    4,
    "-n",
    "--nChips",
    help="Number of chips powered in parallel (e.g. 4 for a quad module, 3 for a triplet, 1 for an SCC.)",
)
OPTIONS["verbosity"]: LogLevel = typer.Option(
    LogLevel.info,
    "-v",
    "--verbosity",
    help="Log level [options: DEBUG, INFO (default) WARNING, ERROR]",
)
OPTIONS["submit"]: bool = typer.Option(
    False,
    help="Submit QC results to: https://docs.google.com/spreadsheets/d/1pw_07F94fg2GJQr8wlvhaRUV63uhsAuBt_S1FEFBzBU/view",
)
OPTIONS["site"]: str = typer.Option(
    "",
    "--site",
    help="Your testing site. Required when submitting results. Please use same short-hand as on production DB, i.e. LBNL_PIXEL_MODULES for LBNL, IRFU for Paris-Saclay, ...",
)
OPTIONS["lp_enable"]: bool = typer.Option(False, help="low power mode")

OPTIONS["input_file"]: Path = typer.Option(
    ...,
    "-i",
    "--input-file",
    help="analysis output file",
    exists=True,
    file_okay=True,
    readable=True,
    resolve_path=True,
)

OPTIONS["config_file"]: Path = typer.Option(
    ...,
    "-c",
    "--config-file",
    help="the config file to be modified",
    exists=True,
    file_okay=True,
    readable=True,
    writable=True,
    resolve_path=True,
)

OPTIONS["input_dir"]: Path = typer.Option(
    ...,
    "-i",
    "--input-dir",
    help="Analysis output directory",
    exists=True,
    dir_okay=True,
    readable=True,
    resolve_path=True,
)
OPTIONS["config_dir"]: Path = typer.Option(
    ...,
    "-c",
    "--config-dir",
    help="Path to the module configuration directory to be modified",
    exists=True,
    dir_okay=True,
    readable=True,
    writable=True,
    resolve_path=True,
)
OPTIONS["config_type"]: str = typer.Option(
    "",
    "-t",
    "--config-type",
    help="The config type to be modified. E.g. E.g. L2_warm/l2_cold.",
)



from __future__ import annotations

import importlib.metadata
import logging
import os
from datetime import datetime
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import typer
from module_qc_data_tools import load_json, outputDataFrame, qcDataFrame, save_dict_list

from module_qc_analysis_tools.cli.globals import (
    CONTEXT_SETTINGS,
    OPTIONS,
    FitMethod,
    LogLevel,
)
from module_qc_analysis_tools.utils.analysis import perform_qc_analysis, submit_results
from module_qc_analysis_tools.utils.misc import (
    DataExtractor,
    JsonChecker,
    bcolors,
    get_inputs,
    get_qc_config,
    get_time_stamp,
    getImuxMap,
    getVmuxMap,
    linear_fit,
    linear_fit_np,
)

app = typer.Typer(context_settings=CONTEXT_SETTINGS)

log = logging.getLogger(__name__)

EMPTY_VAL = -1


def get_NtcCalPar(metadata):
    # Read NTC parameters from metadata in the chip config.
    if "NtcCalPar" in metadata:
        NtcCalPar = metadata.get("NtcCalPar")
    else:
        NtcCalPar = [
            0.0007488999981433153,
            0.0002769000129774213,
            7.059500006789676e-08,
        ]
        log.warning(
            bcolors.WARNING
            + " No NtcCalPar found in the input config file! Using the default NTC parameters."
            + bcolors.ENDC
        )
    return NtcCalPar


def get_NfPar(metadata):
    # Read Nf parameters from metadata in the chip config.
    NfPar = {}
    if "NfASLDO" in metadata:
        NfPar["NfASLDO"] = metadata.get("NfASLDO")
    else:
        NfPar["NfASLDO"] = 1.264
        log.warning(
            bcolors.WARNING
            + " No NfASLDO found in the input config file! Using the default Nf parameter value 1.264."
            + bcolors.ENDC
        )
    if "NfDSLDO" in metadata:
        NfPar["NfDSLDO"] = metadata.get("NfDSLDO")
    else:
        NfPar["NfDSLDO"] = 1.264
        log.warning(
            bcolors.WARNING
            + " No NfASLDO found in the input config file! Using the default Nf parametervalue 1.264."
            + bcolors.ENDC
        )
    if "NfACB" in metadata:
        NfPar["NfACB"] = metadata.get("NfACB")
    else:
        NfPar["NfACB"] = 1.264
        log.warning(
            bcolors.WARNING
            + " No Nfacb found in the input config file! Using the default Nf parameter value 1.264."
            + bcolors.ENDC
        )

    return NfPar


def calculate_T(calculated_data, NtcCalPar, NfPar):
    # Calculate T External NTC
    Vntc = np.array(calculated_data["Vntc"]["Values"])
    Intc = np.array(calculated_data["Intc"]["Values"])

    Rntc = np.mean(Vntc / Intc)
    A = NtcCalPar[0]
    B = NtcCalPar[1]
    C = NtcCalPar[2]
    AR_TEMP_NTC = 1 / (A + B * np.log(Rntc) + C * ((np.log(Rntc)) ** 3)) - 273.15

    log.debug(f" T Ext NTC: {AR_TEMP_NTC} C")

    # Calculate T External External NTC
    AR_TEMP_EXT = np.mean(np.array(calculated_data["TExtExtNTC"]["Values"]))

    log.debug(f" T Ext Ext NTC: {AR_TEMP_EXT} C")

    # Calculate T MOS sensors
    Vmux14 = np.array(calculated_data["VMonSensAna"]["Values"])
    Vmux16 = np.array(calculated_data["VMonSensDig"]["Values"])
    Vmux18 = np.array(calculated_data["VMonSensAcb"]["Values"])

    def calc_temp_sens(Vmux, Nf):
        V_Bias0 = np.mean(Vmux[:16])
        V_Bias1 = np.mean(Vmux[-16:])
        q = 1.602e-19
        kB = 1.38064852e-23
        dV = V_Bias1 - V_Bias0
        T = dV * q / (Nf * kB * np.log(15)) - 273.15
        return T

    AR_TEMP_ASLDO = calc_temp_sens(Vmux14, NfPar["NfASLDO"])
    AR_TEMP_DSLDO = calc_temp_sens(Vmux16, NfPar["NfDSLDO"])
    AR_TEMP_ACB = calc_temp_sens(Vmux18, NfPar["NfACB"])

    log.debug(f" T MonSensAna: {AR_TEMP_ASLDO} C")
    log.debug(f" T MonSensDig: {AR_TEMP_DSLDO} C")
    log.debug(f" T MonSensAcb: {AR_TEMP_ACB} C")

    return (
        round(AR_TEMP_NTC, 1),
        round(AR_TEMP_EXT, 1),
        round(AR_TEMP_ASLDO, 1),
        round(AR_TEMP_DSLDO, 1),
        round(AR_TEMP_ACB, 1),
    )


def round_list(list_values, digit=None):
    rounded_list = []
    for item in list_values:
        if item >= 0.01 or item == 0:
            rounded_list.append(round(item, digit))
        else:
            rounded_list.append(
                float(np.format_float_scientific(item, precision=digit))
            )
    return rounded_list


def plot_vdd_vs_trim(trim, vdd, vdd_name, output_name, chipname, fit_method):
    fig, ax1 = plt.subplots()
    ax1.plot(trim, vdd, "o", label=f"{vdd_name} vs trim")
    if fit_method == "root":
        p1, p0 = linear_fit(trim, vdd)
    elif fit_method == "numpy":
        p1, p0 = linear_fit_np(trim, vdd)
    ax1.axhline(y=1.2, color="r", linestyle="--", label=f"Nominal {vdd_name} value")
    x_line = np.linspace(trim[0], trim[-1], 100)
    ax1.plot(
        x_line,
        p1 * x_line + p0,
        "g-",
        alpha=0.5,
        label=f"Fitted line y = {p1:.3e} * x + {p0:.3e}",
    )
    ax1.set_title(f"{vdd_name} vs Trim Chip {chipname}")
    ax1.set_xlabel("Trim")
    ax1.set_ylabel(f"{vdd_name} (V)")
    ax1.legend()
    log.info(f" Saving {output_name}")
    fig.savefig(output_name)


@app.command()
def main(
    input_meas: Path = OPTIONS["input_meas"],
    base_output_dir: Path = OPTIONS["output_dir"],
    qc_criteria_path: Path = OPTIONS["qc_criteria"],
    layer: str = OPTIONS["layer"],
    permodule: bool = OPTIONS["permodule"],
    submit: bool = OPTIONS["submit"],
    site: str = OPTIONS["site"],
    fit_method: FitMethod = OPTIONS["fit_method"],
    verbosity: LogLevel = OPTIONS["verbosity"],
):
    log.setLevel(verbosity.value)

    if submit and site == "":
        log.error(
            bcolors.ERROR
            + "You have supplied the --submit option without specifying --site (testing site). Please specify your testing site if you would like to submit these results"
            + bcolors.ENDC
        )
        return

    log.info("")
    log.info(" ===============================================")
    log.info(" \tPerforming Analog Readback analysis")
    log.info(" ===============================================")
    log.info("")

    test_type = os.path.basename(__file__).split(".py")[0]
    time_start = round(datetime.timestamp(datetime.now()))
    output_dir = base_output_dir.joinpath(test_type).joinpath(f"{time_start}")
    output_dir.mkdir(parents=True, exist_ok=False)

    allinputs = get_inputs(input_meas, test_type)
    qc_config = get_qc_config(qc_criteria_path, test_type)

    alloutput = []
    timestamps = []

    alloutput_int_biases = []
    timestamps_int_biases = []

    for filename in sorted(allinputs):
        log.info("")
        log.info(f" Loading {filename}")
        meas_timestamp = get_time_stamp(filename)
        inputDFs = load_json(filename)
        log.debug(
            f" There are results from {len(inputDFs)} measuremnet(s) stored in this file"
        )

        chipnames = []
        results = {}
        data = {}
        int_biases = {}
        for inputDF in inputDFs:
            qcframe = inputDF.get_results()
            metadata = qcframe.get_meta_data()

            # Read chipname from input DF
            try:
                chipname = metadata.get("Name")
                chipnames.append(chipname)
                log.debug(f" Found chip name = {chipname} from chip config")
            except Exception:
                log.warning(
                    bcolors.WARNING
                    + "Chip name not found in input from {filename}, skipping."
                    + bcolors.ENDC
                )
                continue

            # Create an output DF for each chip
            if chipname not in data:
                data[chipname] = qcDataFrame()
                data[chipname].add_property(
                    "ANALYSIS_VERSION",
                    importlib.metadata.version("module-qc-analysis-tools"),
                )
                data[chipname]._meta_data.update(metadata)
                int_biases[chipname] = {}

            """"" Check file integrity  """ ""
            checker = JsonChecker(inputDF, test_type)

            try:
                checker.check()
            except BaseException as exc:
                log.exception(exc)
                log.warning(
                    bcolors.WARNING
                    + " JsonChecker check not passed, skipping this input."
                    + bcolors.ENDC
                )
                continue
            else:
                log.debug(" JsonChecker check passed!")
                pass

            """""  Calculate quanties   """ ""
            # Vmux conversion is embedded.
            extractor = DataExtractor(inputDF, test_type)
            calculated_data = extractor.calculate()
            Vmux_map = getVmuxMap()
            Imux_map = getImuxMap()

            AR_values_names = []
            for imux in range(32):
                AR_values_names.append(Imux_map[imux])
            for vmux in range(40):
                AR_values_names.append(Vmux_map[vmux])

            tmpresults = {}
            if inputDF._subtestType == "AR_VMEAS":
                for key in calculated_data:
                    int_biases[chipname][key] = calculated_data[key]["Values"][0]
                AR_values = []
                NOT_MEASURED = 0
                for name in AR_values_names:
                    if name in int_biases[chipname]:
                        AR_values.append(int_biases[chipname][name])
                    else:
                        AR_values.append(NOT_MEASURED)
                data[chipname].add_parameter(
                    "AR_NOMINAL_SETTINGS", round_list(AR_values, 4)
                )
                tmpresults.update({"AR_NOMINAL_SETTINGS": AR_values})

            elif inputDF._subtestType == "AR_TEMP":
                NtcCalPar = get_NtcCalPar(metadata["ChipConfigs"]["RD53B"]["Parameter"])
                NfPar = get_NfPar(metadata["ChipConfigs"]["RD53B"]["Parameter"])
                (
                    AR_TEMP_NTC,
                    AR_TEMP_EXT,
                    AR_TEMP_ASLDO,
                    AR_TEMP_DSLDO,
                    AR_TEMP_ACB,
                ) = calculate_T(calculated_data, NtcCalPar, NfPar)
                # Add parameters for output file
                data[chipname].add_parameter("AR_TEMP_NTC", AR_TEMP_NTC)
                data[chipname].add_parameter("AR_TEMP_EXT", AR_TEMP_EXT)
                data[chipname].add_parameter("AR_TEMP_ASLDO", AR_TEMP_ASLDO)
                data[chipname].add_parameter("AR_TEMP_DSLDO", AR_TEMP_DSLDO)
                data[chipname].add_parameter("AR_TEMP_ACB", AR_TEMP_ACB)
                data[chipname].add_parameter("AR_TEMP_NF_ASLDO", NfPar["NfASLDO"])
                data[chipname].add_parameter("AR_TEMP_NF_DSLDO", NfPar["NfDSLDO"])
                data[chipname].add_parameter("AR_TEMP_NF_ACB", NfPar["NfACB"])
                # Load values to dictionary for QC analysis
                tmpresults.update({"ChipNTC_vs_ExtExt": AR_TEMP_NTC - AR_TEMP_EXT})
                tmpresults.update({"ASLO_ChipNTC": AR_TEMP_ASLDO - AR_TEMP_NTC})
                tmpresults.update({"DSLD_ChipNTC": AR_TEMP_DSLDO - AR_TEMP_NTC})
                tmpresults.update({"ACB_ChipNTC": AR_TEMP_ACB - AR_TEMP_NTC})

            elif inputDF._subtestType == "AR_VDD":
                vdda = calculated_data["VDDA"]["Values"].tolist()
                vddd = calculated_data["VDDD"]["Values"].tolist()
                trimA = calculated_data["SldoTrimA"]["Values"].tolist()
                trimD = calculated_data["SldoTrimD"]["Values"].tolist()
                # Add parameters for output file
                data[chipname].add_parameter("AR_VDDA_VS_TRIM", round_list(vdda, 4))
                data[chipname].add_parameter("AR_VDDD_VS_TRIM", round_list(vddd, 4))
                output_name_vdda = output_dir.joinpath(f"{chipname}_VDDA_TRIM.png")
                output_name_vddd = output_dir.joinpath(f"{chipname}_VDDD_TRIM.png")
                plot_vdd_vs_trim(
                    trimA, vdda, "VDDA", output_name_vdda, chipname, fit_method.value
                )
                plot_vdd_vs_trim(
                    trimD, vddd, "VDDD", output_name_vddd, chipname, fit_method.value
                )
                # Load values to dictionary for QC analysis
                tmpresults.update({"AR_VDDA_VS_TRIM": round_list(vdda, 4)})
                tmpresults.update({"AR_VDDD_VS_TRIM": round_list(vddd, 4)})

            else:
                log.warning(
                    bcolors.WARNING
                    + f"{filename}.json does not have any required subtestType. Skipping."
                    + bcolors.ENDC
                )
                continue

            if results.get(chipname):
                results[chipname].update(tmpresults)
            else:
                results[chipname] = tmpresults

        log.debug(
            f" There are results from {len(chipnames)} chip(s) stored in this file"
        )
        for chip in chipnames:
            # Add not measureed parameters for output file
            data[chip].add_parameter("AR_TEMP_POLY_TOP", EMPTY_VAL)
            data[chip].add_parameter("AR_TEMP_POLY_BOTTOM", EMPTY_VAL)
            data[chip].add_parameter("AR_TEMP_NF_TOP", EMPTY_VAL)
            data[chip].add_parameter("AR_TEMP_NF_BOTTOM", EMPTY_VAL)
            data[chip].add_parameter("AR_RING_OSCILATOR_A", [EMPTY_VAL] * 16)
            data[chip].add_parameter("AR_RING_OSCILATOR_B", [EMPTY_VAL] * 16)

        """"" Output a json file """ ""
        for key, df in data.items():
            outputDF = outputDataFrame()
            outputDF.set_test_type(test_type)
            outputDF.set_results(df)

            # Perform QC analysis
            passes_qc = perform_qc_analysis(
                test_type,
                qc_config,
                layer,
                results.get(key),
                verbosity.value,
            )
            if passes_qc == -1:
                log.error(
                    bcolors.ERROR
                    + f" QC analysis for {key} was NOT successful. Please fix and re-run. Continuing to next chip.."
                    + bcolors.ENDC
                )
                continue
            log.info("")
            if passes_qc:
                log.info(
                    f" Chip {key} passes QC? "
                    + bcolors.OKGREEN
                    + f"{passes_qc}"
                    + bcolors.ENDC
                )
            else:
                log.info(
                    f" Chip {key} passes QC? "
                    + bcolors.BADRED
                    + f"{passes_qc}"
                    + bcolors.ENDC
                )
            log.info("")
            outputDF.set_pass_flag(passes_qc)
            if submit:
                submit_results(
                    outputDF.to_dict(True),
                    time_start,
                    site,
                    output_dir.joinpath("submit.txt"),
                )

            if permodule:
                alloutput += [outputDF.to_dict(True)]
                timestamps += [meas_timestamp]
            else:
                outfile = output_dir.joinpath(f"{key}.json")
                log.info(f" Saving output of analysis to: {outfile}")
                save_dict_list(outfile, [outputDF.to_dict(True)])

        if verbosity.value == "DEBUG":
            # Save an output file for only internal biases
            for key in int_biases:
                if permodule:
                    alloutput_int_biases += [int_biases[key]]
                    timestamps_int_biases += [meas_timestamp]
                else:
                    outfile = output_dir.joinpath(f"{key}_internal_biases.json")
                    log.info(f" Saving DEBUG file with internal biases to: {outfile}")
                    save_dict_list(outfile, [int_biases[key]])
    if permodule:
        # Only store results from same timestamp into same file
        dfs = np.array(alloutput)
        tss = np.array(timestamps)
        for x in np.unique(tss):
            outfile = output_dir.joinpath("module.json")
            log.info(f" Saving output of analysis to: {outfile}")
            save_dict_list(
                outfile,
                dfs[tss == x].tolist(),
            )
        if verbosity.value == "DEBUG":
            # Save an output file for only internal biases
            dfs = np.array(alloutput_int_biases)
            tss = np.array(timestamps_int_biases)
            for x in np.unique(tss):
                outfile = output_dir.joinpath(f"internal_biases_{x}.json")
                log.info(f" Saving DEBUG file with internal biases to: {outfile}")
                save_dict_list(
                    outfile,
                    dfs[tss == x].tolist(),
                )


if __name__ == "__main__":
    typer.run(main)



from __future__ import annotations

import json
import logging
from pathlib import Path

import typer

from module_qc_analysis_tools.cli.globals import CONTEXT_SETTINGS, OPTIONS

log = logging.getLogger(__name__)

stack = []


def lookup(input_dict, search_key):
    # recursion to look up the desired key in a dict and record the path
    for k, v in input_dict.items():
        if k == search_key:
            return v
        elif isinstance(v, dict):
            _v = lookup(v, search_key)
            if _v is not None:
                stack.append(k)
                return _v


def overwrite(in_file, config_file):
    with config_file.open() as jsonFile:
        config_file_data = json.load(jsonFile)
    with in_file.open() as jsonFile:
        in_file_data = json.load(jsonFile)

    original_par = lookup(config_file_data, "Parameter")

    if original_par is None:
        raise KeyError(f"Parameter not found in config file {config_file}! ")

    # Check if chip name matched
    if in_file_data["Identifiers"]["Name"] != original_par["Name"]:
        raise KeyError("Chip name mismatched! Please check the input")

    # overwrite the Parameter part
    for k, v in in_file_data.items():
        if k in original_par.keys():
            log.info(f"[{k}] Change from {original_par[k]} to {v}")
            original_par[k] = v

    stack.reverse()

    part_config_file_data = config_file_data
    for k in stack:
        part_config_file_data = part_config_file_data[k]
    part_config_file_data["Parameter"] = original_par

    with config_file.open("w") as jsonFile:
        json.dump(config_file_data, jsonFile, indent=4)


app = typer.Typer(context_settings=CONTEXT_SETTINGS)


@app.command()
def main(
    input_file: Path = OPTIONS["input_file"], config_file: Path = OPTIONS["config_file"]
):
    overwrite(input_file, config_file)


if __name__ == "__main__":
    typer.run(main)



from __future__ import annotations

import json
import logging
from pathlib import Path

import numpy as np
import typer
from module_qc_data_tools import convert_name_to_serial

from module_qc_analysis_tools.cli.globals import CONTEXT_SETTINGS, OPTIONS
from module_qc_analysis_tools.utils.misc import get_inputs

log = logging.getLogger(__name__)
log.setLevel("INFO")


class WriteConfig:
    ##############
    # This class converts a parameter in a chip config to a given value.
    # `in_path` must be the path to the output directory of the analysis.
    # `config_path` must be the path to the directory ofchip config file in Yarr.
    # If `permodule`, the paths provided much be the path of the correct directory.
    ##############
    def __init__(
        self,
        in_path: Path | None = None,
        config_path: Path | None = None,
        config_type: str | None = None,
    ):
        self.in_path = in_path or Path()
        _config_path = config_path or Path()
        self.config_path = (
            _config_path.joinpath(config_type) if config_type else _config_path
        )
        self.stack = []
        self.all_test_types = [
            "ADC_CALIBRATION",
            "ANALOG_READBACK",
            "VCAL_CALIBRATION",
            "INJECTION_CAPACITANCE",
        ]
        self.config_chip_name = ""
        self.in_files = get_inputs(self.in_path, self.get_test_type())
        self.config_files = get_inputs(self.config_path, self.get_test_type(), True)

    def get_test_type(self):
        for test_type in self.all_test_types:
            if test_type in str(self.in_path):
                return test_type

    def reset_stack(self):
        self.stack = []

    def lookup(self, input_dict, search_key, update_stack=False):
        # recursion to look up the desired key in a dict and record the path
        for k, v in input_dict.items():
            if k == search_key:
                return v
            elif isinstance(v, dict):
                _v = self.lookup(v, search_key, update_stack)
                if _v is not None:
                    if update_stack:
                        self.stack.append(k)
                    return _v

    def set_ADCcalPar(self, in_file, config_file):
        # Set the calibrated ADC parameters from the analysis.
        ADC_CALIBRATION_SLOPE = float(self.lookup(in_file, "ADC_CALIBRATION_SLOPE"))
        ADC_CALIBRATION_OFFSET = float(self.lookup(in_file, "ADC_CALIBRATION_OFFSET"))
        self.overwrite(config_file, "ADCcalPar", ADC_CALIBRATION_SLOPE, index=1)
        self.overwrite(config_file, "ADCcalPar", ADC_CALIBRATION_OFFSET, index=0)

    def set_trim(self, in_file, config_file):
        # Set the trim values that gives the closest to nominal vdd value.
        vdda = np.array(self.lookup(in_file, "AR_VDDA_VS_TRIM"))
        vddd = np.array(self.lookup(in_file, "AR_VDDD_VS_TRIM"))
        SldoTrimA = int(np.absolute(vdda - 1.2).argmin())
        SldoTrimD = int(np.absolute(vddd - 1.2).argmin())
        self.overwrite(config_file, "SldoTrimA", SldoTrimA)
        self.overwrite(config_file, "SldoTrimD", SldoTrimD)

    def set_VcalPar(self, in_file, config_file):
        # Set the calibrated VCAL parameters from the analysis.
        VCAL_MED_SLOPE = float(self.lookup(in_file, "VCAL_MED_SLOPE"))
        VCAL_MED_OFFSET = float(self.lookup(in_file, "VCAL_MED_OFFSET"))
        self.overwrite(config_file, "VcalPar", [VCAL_MED_SLOPE, VCAL_MED_OFFSET])

    def set_InjCap(self, in_file, config_file):
        # Set the Injection Capacitance value from the analysis.
        INJ_CAPACITANC = float(self.lookup(in_file, "INJ_CAPACITANCE"))
        self.overwrite(config_file, "InjCap", INJ_CAPACITANC)

    def overwrite(self, config_file, search_key, set_value, index=None):
        # search_key is a string which is the name of the parameter that will be overwritten.
        # set_value is the value the parameter that will be overwritten to.
        with config_file.open() as jsonFile:
            config_file_data = json.load(jsonFile)
        self.reset_stack()
        original_search_key = self.lookup(config_file_data, search_key, True)
        if original_search_key is None:
            raise KeyError(f"Parameter not found in config file {config_file}! ")
        log.info(
            f"Chip {self.config_chip_name} [{search_key}] change from {original_search_key} to {set_value}."
        )
        if index is None:
            original_search_key = set_value
        else:
            original_search_key[index] = set_value

        self.stack.reverse()

        part_config_file_data = config_file_data
        for k in self.stack:
            part_config_file_data = part_config_file_data[k]
        part_config_file_data[search_key] = original_search_key

        with config_file.open("w") as jsonFile:
            json.dump(config_file_data, jsonFile, indent=4)

    def update_config(self):
        for config_file in self.config_files:
            with config_file.open() as jsonFile:
                config_file_data = json.load(jsonFile)

            out_file = self.in_path.joinpath(f"{config_file.name}.before")
            if out_file.exists():
                log.warning(f"File {out_file} already exists! Skip overwriting!")
            else:
                with open(out_file, "w") as fp:
                    json.dump(config_file_data, fp, indent=4)
            self.config_chip_name = self.lookup(config_file_data, "Name")
            config_chip_serial = convert_name_to_serial(self.config_chip_name)

            # overwrite the parameter
            for in_file in self.in_files:
                found_chip = False
                with in_file.open() as jsonFile:
                    in_file_data = json.load(jsonFile)
                for chip_data in in_file_data:
                    # Check if chip name matched
                    in_chip_serial = self.lookup(chip_data, "serialNumber")
                    if in_chip_serial is None:
                        raise KeyError(
                            f"Chip {self.config_chip_name} not found in the input files! Please check the input files."
                        )
                    if in_chip_serial != config_chip_serial:
                        log.debug(
                            f"Chip {self.config_chip_name} not found in config. Checking the next chip."
                        )
                    else:
                        found_chip = True
                        test_type = self.get_test_type()
                        if test_type == "ADC_CALIBRATION":
                            self.set_ADCcalPar(chip_data, config_file)
                        elif test_type == "ANALOG_READBACK":
                            self.set_trim(chip_data, config_file)
                        elif test_type == "VCAL_CALIBRATION":
                            self.set_VcalPar(chip_data, config_file)
                        elif test_type == "INJECTION_CAPACITANCE":
                            self.set_InjCap(chip_data, config_file)
                        else:
                            raise KeyError(
                                f"No test type found in the input directory {self.input_path}! Please check the input path."
                            )
                        break
                if found_chip:
                    break
            with open(config_file) as jsonFile:
                config_file_data = json.load(jsonFile)

            out_file = self.in_path.joinpath(f"{config_file.name}.after")
            with open(out_file, "w") as fp:
                json.dump(config_file_data, fp, indent=4)
            if not found_chip:
                raise KeyError(
                    f"Chip {self.config_chip_name} with serial number {config_chip_serial} not found! Please check the input files."
                )


app = typer.Typer(context_settings=CONTEXT_SETTINGS)


@app.command()
def main(
    input_dir: Path = OPTIONS["input_dir"],
    config_dir: Path = OPTIONS["config_dir"],
    config_type: str = OPTIONS["config_type"],
):
    log.info(" ==========================================")
    wc = WriteConfig(input_dir, config_dir, config_type)
    wc.update_config()


if __name__ == "__main__":
    typer.run(main)
